from __future__ import annotations

from typing import TYPE_CHECKING, Optional, TypeVar

ResponseClass = TypeVar("ResponseClass")

if TYPE_CHECKING:
    from infrahub.message_bus import InfrahubMessage
    from infrahub.message_bus.types import MessageTTL
    from infrahub.services import InfrahubServices


class InfrahubMessageBus:
    DELIVER_TIMEOUT: int = 30 * 60  # 30 minutes
    worker_bindings: list[str] = [
        "check.*.*",
        "event.*.*",
        "finalize.*.*",
        "git.*.*",
        "refresh.webhook.*",
        "request.*.*",
        "send.*.*",
        "schema.*.*",
        "transform.*.*",
        "trigger.*.*",
    ]
    event_bindings: list[str] = ["refresh.registry.*"]
    broadcasted_event_bindings: list[str] = ["refresh.git.*"]

    async def initialize(self, service: InfrahubServices) -> None:
        """Initialize the Message bus"""

    async def shutdown(self) -> None:
        """Shutdown the Message bus"""

    async def publish(
        self, message: InfrahubMessage, routing_key: str, delay: Optional[MessageTTL] = None, is_retry: bool = False
    ) -> None:
        raise NotImplementedError()

    async def reply(self, message: InfrahubMessage, routing_key: str) -> None:
        raise NotImplementedError()

    async def rpc(self, message: InfrahubMessage, response_class: type[ResponseClass]) -> ResponseClass:
        raise NotImplementedError()



import importlib
import logging
import os
from typing import Any, Optional

import typer
from anyio.abc import TaskStatus
from infrahub_sdk import Config, InfrahubClient
from infrahub_sdk.exceptions import Error as SdkError
from prefect import settings as prefect_settings
from prefect.client.schemas.objects import FlowRun
from prefect.flow_engine import run_flow_async
from prefect.logging.handlers import APILogHandler
from prefect.workers.base import BaseJobConfiguration, BaseVariables, BaseWorker, BaseWorkerResult
from prometheus_client import start_http_server

from infrahub import __version__ as infrahub_version
from infrahub import config
from infrahub.components import ComponentType
from infrahub.core import registry
from infrahub.core.initialization import initialization
from infrahub.database import InfrahubDatabase, get_db
from infrahub.dependencies.registry import build_component_registry
from infrahub.git import initialize_repositories_directory
from infrahub.lock import initialize_lock
from infrahub.services import InfrahubServices, services
from infrahub.services.adapters.cache import InfrahubCache
from infrahub.services.adapters.cache.nats import NATSCache
from infrahub.services.adapters.cache.redis import RedisCache
from infrahub.services.adapters.message_bus import InfrahubMessageBus
from infrahub.services.adapters.message_bus.nats import NATSMessageBus
from infrahub.services.adapters.message_bus.rabbitmq import RabbitMQMessageBus
from infrahub.services.adapters.workflow import InfrahubWorkflow
from infrahub.services.adapters.workflow.local import WorkflowLocalExecution
from infrahub.services.adapters.workflow.worker import WorkflowWorkerExecution
from infrahub.workflows.models import TASK_RESULT_STORAGE_NAME

WORKER_QUERY_SECONDS = "2"
WORKER_DEFAULT_RESULT_STORAGE_BLOCK = f"redisstoragecontainer/{TASK_RESULT_STORAGE_NAME}"
DEFAULT_TASK_LOGGERS = ["infrahub.tasks"]


class InfrahubWorkerAsyncConfiguration(BaseJobConfiguration):
    env: dict[str, str | None] = {
        "PREFECT_WORKER_QUERY_SECONDS": WORKER_QUERY_SECONDS,
        "PREFECT_DEFAULT_RESULT_STORAGE_BLOCK": WORKER_DEFAULT_RESULT_STORAGE_BLOCK,
    }
    labels: dict[str, str] = {
        "infrahub.app/version": infrahub_version,
    }


class InfrahubWorkerAsyncTemplateVariables(BaseVariables):
    pass


class InfrahubWorkerAsyncResult(BaseWorkerResult):
    """Result returned by the InfrahubWorker."""


class InfrahubWorkerAsync(BaseWorker):
    type: str = "infrahubasync"
    job_configuration = InfrahubWorkerAsyncConfiguration
    job_configuration_variables = InfrahubWorkerAsyncTemplateVariables
    _documentation_url = "https://example.com/docs"
    _logo_url = "https://example.com/logo"
    _description = "Infrahub worker designed to run the flow in the main async loop."

    async def setup(
        self,
        client: InfrahubClient | None = None,
        metric_port: int | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        logging.getLogger("websockets").setLevel(logging.ERROR)
        logging.getLogger("httpx").setLevel(logging.ERROR)
        logging.getLogger("httpcore").setLevel(logging.ERROR)
        logging.getLogger("neo4j").setLevel(logging.ERROR)
        logging.getLogger("aio_pika").setLevel(logging.ERROR)
        logging.getLogger("aiormq").setLevel(logging.ERROR)
        logging.getLogger("git").setLevel(logging.ERROR)
        # Prevent git from interactively prompting the user for passwords if the credentials provided
        # by the credential helper is failing.
        os.environ["GIT_TERMINAL_PROMPT"] = "0"

        if not config.SETTINGS.settings:
            config_file = os.environ.get("INFRAHUB_CONFIG", "infrahub.toml")
            config.load_and_exit(config_file_name=config_file)

        self._init_logger()

        # Start metric endpoint
        if metric_port is None or metric_port != 0:
            metric_port = metric_port or int(os.environ.get("INFRAHUB_METRICS_PORT", 8000))
            self._logger.info(f"Starting metric endpoint on port {metric_port}")
            start_http_server(metric_port)

        await super().setup(**kwargs)

        self._exit_stack.enter_context(
            prefect_settings.temporary_settings(
                updates={  # type: ignore[arg-type]
                    prefect_settings.PREFECT_WORKER_QUERY_SECONDS: config.SETTINGS.workflow.worker_polling_interval,
                    prefect_settings.PREFECT_RESULTS_PERSIST_BY_DEFAULT: True,
                    prefect_settings.PREFECT_DEFAULT_RESULT_STORAGE_BLOCK: WORKER_DEFAULT_RESULT_STORAGE_BLOCK,
                }
            )
        )

        client = await self._init_infrahub_client(client=client)
        service = await self._init_services(client=client)

        if not registry.schema_has_been_initialized():
            initialize_lock(service=service)

            async with service.database.start_session() as db:
                await initialization(db=db)

            await service.component.refresh_schema_hash()

        initialize_repositories_directory()
        build_component_registry()
        await service.scheduler.start_schedule()
        self._logger.info("Worker initialization completed .. ")

    async def run(
        self,
        flow_run: FlowRun,
        configuration: BaseJobConfiguration,
        task_status: Optional[TaskStatus] = None,
    ) -> BaseWorkerResult:
        flow_run_logger = self.get_flow_run_logger(flow_run)

        entrypoint: str = configuration._related_objects["deployment"].entrypoint

        file_path, flow_name = entrypoint.split(":")
        file_path.replace("/", ".")
        module_path = file_path.replace("backend/", "").replace(".py", "").replace("/", ".")
        module = importlib.import_module(module_path)
        flow_func = getattr(module, flow_name)

        flow_run_logger.debug("Validating parameters")
        params = flow_func.validate_parameters(parameters=flow_run.parameters)

        if task_status:
            task_status.started()

        await run_flow_async(flow=flow_func, flow_run=flow_run, parameters=params, return_type="state")

        return InfrahubWorkerAsyncResult(
            status_code=0,
            identifier=str(flow_run.id),
        )

    def _init_logger(self) -> None:
        """Initialize loggers to use the API handle provided by Prefect."""
        api_handler = APILogHandler()

        for logger_name in config.SETTINGS.workflow.extra_loggers + DEFAULT_TASK_LOGGERS:
            logger = logging.getLogger(logger_name)
            logger.setLevel(config.SETTINGS.workflow.extra_log_level.value)
            logger.addHandler(api_handler)

    async def _init_infrahub_client(self, client: InfrahubClient | None = None) -> InfrahubClient:
        if not client:
            self._logger.debug(f"Using Infrahub API at {config.SETTINGS.main.internal_address}")
            client = InfrahubClient(
                config=Config(address=config.SETTINGS.main.internal_address, retry_on_failure=True, log=self._logger)
            )

        try:
            await client.branch.all()
        except SdkError as exc:
            self._logger.error(f"Error in communication with Infrahub: {exc.message}")
            raise typer.Exit(1)

        return client

    async def _init_database(self) -> InfrahubDatabase:
        return InfrahubDatabase(driver=await get_db(retry=1))

    async def _init_workflow(self) -> InfrahubWorkflow:
        return config.OVERRIDE.workflow or (
            WorkflowWorkerExecution()
            if config.SETTINGS.workflow.driver == config.WorkflowDriver.WORKER
            else WorkflowLocalExecution()
        )

    async def _init_message_bus(self) -> InfrahubMessageBus:
        return config.OVERRIDE.message_bus or (
            NATSMessageBus() if config.SETTINGS.broker.driver == config.BrokerDriver.NATS else RabbitMQMessageBus()
        )

    async def _init_cache(self) -> InfrahubCache:
        return config.OVERRIDE.cache or (
            NATSCache() if config.SETTINGS.cache.driver == config.CacheDriver.NATS else RedisCache()
        )

    async def _init_services(self, client: InfrahubClient) -> InfrahubServices:
        database = await self._init_database()
        workflow = await self._init_workflow()
        message_bus = await self._init_message_bus()
        cache = await self._init_cache()

        service = InfrahubServices(
            cache=cache,
            client=client,
            database=database,
            message_bus=message_bus,
            workflow=workflow,
            component_type=ComponentType.GIT_AGENT,
        )

        services.service = service
        await service.initialize()

        return service



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Iterable, Literal, Sequence, overload
from uuid import uuid4

from infrahub import lock
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import ValidationError
from infrahub.log import get_logger

from .model.path import (
    BranchTrackingId,
    EnrichedDiffRoot,
    EnrichedDiffRootMetadata,
    EnrichedDiffs,
    EnrichedDiffsMetadata,
    NameTrackingId,
    TrackingId,
)

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node

    from .calculator import DiffCalculator
    from .combiner import DiffCombiner
    from .conflict_transferer import DiffConflictTransferer
    from .conflicts_enricher import ConflictsEnricher
    from .data_check_synchronizer import DiffDataCheckSynchronizer
    from .enricher.aggregated import AggregatedDiffEnricher
    from .enricher.labels import DiffLabelsEnricher
    from .repository.repository import DiffRepository


log = get_logger()


@dataclass
class EnrichedDiffRequest:
    base_branch: Branch
    diff_branch: Branch
    from_time: Timestamp
    to_time: Timestamp
    tracking_id: TrackingId
    node_field_specifiers: dict[str, set[str]] = field(default_factory=dict)

    def __repr__(self) -> str:
        return (
            f"EnrichedDiffRequest(base_branch_name={self.base_branch.name}, diff_branch_name={self.diff_branch.name},"
            f" from_time={self.from_time.to_string()}, to_time={self.to_time.to_string()},"
            f" tracking_id={self.tracking_id.serialize() if self.tracking_id else None}),"
            f" num_node_field_specifiers={len(self.node_field_specifiers)}"
        )


class DiffCoordinator:
    lock_namespace = "diff-update"

    def __init__(
        self,
        diff_repo: DiffRepository,
        diff_calculator: DiffCalculator,
        diff_enricher: AggregatedDiffEnricher,
        diff_combiner: DiffCombiner,
        conflicts_enricher: ConflictsEnricher,
        labels_enricher: DiffLabelsEnricher,
        data_check_synchronizer: DiffDataCheckSynchronizer,
        conflict_transferer: DiffConflictTransferer,
    ) -> None:
        self.diff_repo = diff_repo
        self.diff_calculator = diff_calculator
        self.diff_enricher = diff_enricher
        self.diff_combiner = diff_combiner
        self.conflicts_enricher = conflicts_enricher
        self.labels_enricher = labels_enricher
        self.data_check_synchronizer = data_check_synchronizer
        self.conflict_transferer = conflict_transferer
        self.lock_registry = lock.registry

    async def run_update(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: str | None = None,
        to_time: str | None = None,
        name: str | None = None,
    ) -> None:
        # we are updating a diff that tracks the full lifetime of a branch
        if not name and not from_time and not to_time:
            await self.update_branch_diff(base_branch=base_branch, diff_branch=diff_branch)
            return

        if from_time:
            from_timestamp = Timestamp(from_time)
        else:
            from_timestamp = Timestamp(diff_branch.get_branched_from())
        if to_time:
            to_timestamp = Timestamp(to_time)
        else:
            to_timestamp = Timestamp()
        if not name:
            raise ValidationError("diff with specified time range requires a name")
        await self.create_or_update_arbitrary_timeframe_diff(
            base_branch=base_branch,
            diff_branch=diff_branch,
            from_time=from_timestamp,
            to_time=to_timestamp,
            name=name,
        )

    def _get_lock_name(self, base_branch_name: str, diff_branch_name: str, is_incremental: bool) -> str:
        lock_name = f"{base_branch_name}__{diff_branch_name}"
        if is_incremental:
            lock_name += "__incremental"
        return lock_name

    async def update_branch_diff(self, base_branch: Branch, diff_branch: Branch) -> EnrichedDiffRootMetadata:
        log.info(f"Received request to update branch diff for {base_branch.name} - {diff_branch.name}")
        incremental_lock_name = self._get_lock_name(
            base_branch_name=base_branch.name, diff_branch_name=diff_branch.name, is_incremental=True
        )
        existing_incremental_lock = self.lock_registry.get_existing(
            name=incremental_lock_name, namespace=self.lock_namespace
        )
        if existing_incremental_lock and await existing_incremental_lock.locked():
            log.info(f"Branch diff update for {base_branch.name} - {diff_branch.name} already in progress")
            async with self.lock_registry.get(name=incremental_lock_name, namespace=self.lock_namespace):
                log.info(f"Existing branch diff update for {base_branch.name} - {diff_branch.name} complete")
                return await self.diff_repo.get_one(
                    tracking_id=BranchTrackingId(name=diff_branch.name), diff_branch_name=diff_branch.name
                )
        general_lock_name = self._get_lock_name(
            base_branch_name=base_branch.name, diff_branch_name=diff_branch.name, is_incremental=False
        )
        from_time = Timestamp(diff_branch.get_branched_from())
        to_time = Timestamp()
        tracking_id = BranchTrackingId(name=diff_branch.name)
        async with (
            self.lock_registry.get(name=general_lock_name, namespace=self.lock_namespace),
            self.lock_registry.get(name=incremental_lock_name, namespace=self.lock_namespace),
        ):
            log.info(f"Acquired lock to run branch diff update for {base_branch.name} - {diff_branch.name}")
            enriched_diffs = await self._update_diffs(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=from_time,
                to_time=to_time,
                tracking_id=tracking_id,
                force_branch_refresh=False,
            )
            await self.diff_repo.save(enriched_diffs=enriched_diffs)
            await self._update_core_data_checks(enriched_diff=enriched_diffs.diff_branch_diff)
            log.info(f"Branch diff update complete for {base_branch.name} - {diff_branch.name}")
        return enriched_diffs.diff_branch_diff

    async def create_or_update_arbitrary_timeframe_diff(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        name: str,
    ) -> EnrichedDiffRootMetadata:
        tracking_id = NameTrackingId(name=name)
        general_lock_name = self._get_lock_name(
            base_branch_name=base_branch.name, diff_branch_name=diff_branch.name, is_incremental=False
        )
        async with self.lock_registry.get(name=general_lock_name, namespace=self.lock_namespace):
            log.info(f"Acquired lock to run arbitrary diff update for {base_branch.name} - {diff_branch.name}")
            enriched_diffs = await self._update_diffs(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=from_time,
                to_time=to_time,
                tracking_id=tracking_id,
                force_branch_refresh=False,
            )

            await self.diff_repo.save(enriched_diffs=enriched_diffs)
            await self._update_core_data_checks(enriched_diff=enriched_diffs.diff_branch_diff)
            log.info(f"Arbitrary diff update complete for {base_branch.name} - {diff_branch.name}")
        return enriched_diffs.diff_branch_diff

    async def recalculate(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        diff_id: str,
    ) -> EnrichedDiffRoot:
        general_lock_name = self._get_lock_name(
            base_branch_name=base_branch.name, diff_branch_name=diff_branch.name, is_incremental=False
        )
        async with self.lock_registry.get(name=general_lock_name, namespace=self.lock_namespace):
            log.info(f"Acquired lock to recalculate diff for {base_branch.name} - {diff_branch.name}")
            current_branch_diff = await self.diff_repo.get_one(diff_branch_name=diff_branch.name, diff_id=diff_id)
            current_base_diff = await self.diff_repo.get_one(
                diff_branch_name=base_branch.name, diff_id=current_branch_diff.partner_uuid
            )
            if current_branch_diff.tracking_id and isinstance(current_branch_diff.tracking_id, BranchTrackingId):
                to_time = Timestamp()
            else:
                to_time = current_branch_diff.to_time
            await self.diff_repo.delete_diff_roots(diff_root_uuids=[current_branch_diff.uuid, current_base_diff.uuid])
            from_time = current_branch_diff.from_time
            branched_from_time = Timestamp(diff_branch.get_branched_from())
            from_time = max(from_time, branched_from_time)
            enriched_diffs = await self._update_diffs(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=branched_from_time,
                to_time=to_time,
                tracking_id=current_branch_diff.tracking_id,
                force_branch_refresh=True,
            )
            if current_branch_diff:
                await self.conflict_transferer.transfer(
                    earlier=current_branch_diff, later=enriched_diffs.diff_branch_diff
                )

            await self.diff_repo.save(enriched_diffs=enriched_diffs)
            await self._update_core_data_checks(enriched_diff=enriched_diffs.diff_branch_diff)
            log.info(f"Diff recalculation complete for {base_branch.name} - {diff_branch.name}")
        return enriched_diffs.diff_branch_diff

    def _get_ordered_diff_pairs(
        self, diff_pairs: Iterable[EnrichedDiffsMetadata], allow_overlap: bool = False
    ) -> list[EnrichedDiffsMetadata]:
        ordered_diffs = sorted(diff_pairs, key=lambda d: d.diff_branch_diff.from_time)
        if allow_overlap:
            return ordered_diffs
        ordered_diffs_no_overlaps: list[EnrichedDiffsMetadata] = []
        for candidate_diff_pair in ordered_diffs:
            if not ordered_diffs_no_overlaps:
                ordered_diffs_no_overlaps.append(candidate_diff_pair)
                continue
            # no time overlap
            previous_diff = ordered_diffs_no_overlaps[-1].diff_branch_diff
            candidate_diff = candidate_diff_pair.diff_branch_diff
            if previous_diff.to_time <= candidate_diff.from_time:
                ordered_diffs_no_overlaps.append(candidate_diff_pair)
                continue
            previous_interval = previous_diff.time_range
            candidate_interval = candidate_diff.time_range
            # keep the diff that covers the larger time frame
            if candidate_interval > previous_interval:
                ordered_diffs_no_overlaps[-1] = candidate_diff_pair
        return ordered_diffs_no_overlaps

    def _build_enriched_diffs_with_no_nodes(self, diff_request: EnrichedDiffRequest) -> EnrichedDiffs:
        base_uuid = str(uuid4())
        branch_uuid = str(uuid4())
        return EnrichedDiffs(
            base_branch_name=diff_request.base_branch.name,
            diff_branch_name=diff_request.diff_branch.name,
            base_branch_diff=EnrichedDiffRoot(
                base_branch_name=diff_request.base_branch.name,
                diff_branch_name=diff_request.base_branch.name,
                from_time=diff_request.from_time,
                to_time=diff_request.to_time,
                tracking_id=diff_request.tracking_id,
                uuid=base_uuid,
                partner_uuid=branch_uuid,
            ),
            diff_branch_diff=EnrichedDiffRoot(
                base_branch_name=diff_request.base_branch.name,
                diff_branch_name=diff_request.diff_branch.name,
                from_time=diff_request.from_time,
                to_time=diff_request.to_time,
                tracking_id=diff_request.tracking_id,
                uuid=branch_uuid,
                partner_uuid=base_uuid,
            ),
        )

    @overload
    async def _update_diffs(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        tracking_id: TrackingId,
        force_branch_refresh: Literal[True] = ...,
    ) -> EnrichedDiffs: ...

    @overload
    async def _update_diffs(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        tracking_id: TrackingId,
        force_branch_refresh: Literal[False] = ...,
    ) -> EnrichedDiffs | EnrichedDiffsMetadata: ...

    async def _update_diffs(
        self,
        base_branch: Branch,
        diff_branch: Branch,
        from_time: Timestamp,
        to_time: Timestamp,
        tracking_id: TrackingId,
        force_branch_refresh: bool = False,
    ) -> EnrichedDiffs | EnrichedDiffsMetadata:
        # start with empty diffs b/c we only care about their metadata for now, hydrate them with data as needed
        diff_pairs_metadata = await self.diff_repo.get_diff_pairs_metadata(
            base_branch_names=[base_branch.name],
            diff_branch_names=[diff_branch.name],
            from_time=from_time,
            to_time=to_time,
            tracking_id=tracking_id,
        )
        aggregated_enriched_diffs = await self._aggregate_enriched_diffs(
            diff_request=EnrichedDiffRequest(
                base_branch=base_branch,
                diff_branch=diff_branch,
                from_time=from_time,
                to_time=to_time,
                tracking_id=tracking_id,
            ),
            partial_enriched_diffs=diff_pairs_metadata if not force_branch_refresh else None,
        )
        diff_uuids_to_delete: list[str] = []
        for diff_pair in diff_pairs_metadata:
            if (
                diff_pair.base_branch_diff.tracking_id == tracking_id
                and diff_pair.base_branch_diff.uuid != aggregated_enriched_diffs.base_branch_diff.uuid
                and diff_pair.base_branch_diff.exists_on_database
            ):
                diff_uuids_to_delete.append(diff_pair.base_branch_diff.uuid)
            if (
                diff_pair.diff_branch_diff.tracking_id == tracking_id
                and diff_pair.diff_branch_diff.uuid != aggregated_enriched_diffs.diff_branch_diff.uuid
                and diff_pair.diff_branch_diff.exists_on_database
            ):
                diff_uuids_to_delete.append(diff_pair.diff_branch_diff.uuid)

        if diff_uuids_to_delete:
            await self.diff_repo.delete_diff_roots(diff_root_uuids=diff_uuids_to_delete)

        # this is an EnrichedDiffsMetadata, so there are no nodes to enrich
        if not isinstance(aggregated_enriched_diffs, EnrichedDiffs):
            aggregated_enriched_diffs.update_metadata(from_time=from_time, to_time=to_time, tracking_id=tracking_id)
            return aggregated_enriched_diffs

        await self.conflicts_enricher.add_conflicts_to_branch_diff(
            base_diff_root=aggregated_enriched_diffs.base_branch_diff,
            branch_diff_root=aggregated_enriched_diffs.diff_branch_diff,
        )
        await self.labels_enricher.enrich(
            enriched_diff_root=aggregated_enriched_diffs.diff_branch_diff, conflicts_only=True
        )

        return aggregated_enriched_diffs

    @overload
    async def _aggregate_enriched_diffs(
        self,
        diff_request: EnrichedDiffRequest,
        partial_enriched_diffs: list[EnrichedDiffsMetadata],
    ) -> EnrichedDiffs | EnrichedDiffsMetadata: ...

    @overload
    async def _aggregate_enriched_diffs(
        self,
        diff_request: EnrichedDiffRequest,
        partial_enriched_diffs: None,
    ) -> EnrichedDiffs: ...

    async def _aggregate_enriched_diffs(
        self,
        diff_request: EnrichedDiffRequest,
        partial_enriched_diffs: list[EnrichedDiffsMetadata] | None,
    ) -> EnrichedDiffs | EnrichedDiffsMetadata:
        """
        If return is an EnrichedDiffsMetadata, it acts as a pointer to a diff in the database that has all the
            necessary data for this diff_request. Might have a different time range and/or tracking_id
        """
        aggregated_enriched_diffs: EnrichedDiffs | EnrichedDiffsMetadata | None = None
        if not partial_enriched_diffs:
            # no existing diffs to use in calculating this diff, so calculate the whole thing and return it
            aggregated_enriched_diffs = await self._calculate_enriched_diff(
                diff_request=diff_request, is_incremental_diff=False
            )

        if partial_enriched_diffs is not None and not aggregated_enriched_diffs:
            ordered_diffs = self._get_ordered_diff_pairs(diff_pairs=partial_enriched_diffs, allow_overlap=False)
            ordered_diff_reprs = [repr(d) for d in ordered_diffs]
            log.info(f"Ordered diffs for aggregation: {ordered_diff_reprs}")
            incremental_diffs_and_requests: list[EnrichedDiffsMetadata | EnrichedDiffRequest | None] = []
            current_time = diff_request.from_time
            while current_time < diff_request.to_time:
                # the next diff to include has already been calculated
                if ordered_diffs and ordered_diffs[0].diff_branch_diff.from_time == current_time:
                    current_diff = ordered_diffs.pop(0)
                    incremental_diffs_and_requests.append(current_diff)
                    current_time = current_diff.diff_branch_diff.to_time
                    continue
                # set the end time to the start of the next calculated diff or the end of the time range
                if ordered_diffs:
                    end_time = ordered_diffs[0].diff_branch_diff.from_time
                else:
                    end_time = diff_request.to_time
                # if there are no changes on either branch in this time range, then there cannot be a diff
                log.info(
                    f"Checking number of changes on branches for {diff_request!r}, from_time={current_time}, to_time={end_time}"
                )
                num_changes_by_branch = await self.diff_repo.get_num_changes_in_time_range_by_branch(
                    branch_names=[diff_request.base_branch.name, diff_request.diff_branch.name],
                    from_time=current_time,
                    to_time=end_time,
                )
                log.info(f"Number of changes: {num_changes_by_branch}")
                might_have_changes_in_time_range = any(num_changes_by_branch.values())
                if not might_have_changes_in_time_range:
                    incremental_diffs_and_requests.append(None)
                    current_time = end_time
                    continue

                incremental_diffs_and_requests.append(
                    EnrichedDiffRequest(
                        base_branch=diff_request.base_branch,
                        diff_branch=diff_request.diff_branch,
                        from_time=current_time,
                        to_time=end_time,
                        tracking_id=diff_request.tracking_id,
                    )
                )
                current_time = end_time

            aggregated_enriched_diffs = await self._concatenate_diffs_and_requests(
                diff_or_request_list=incremental_diffs_and_requests, full_diff_request=diff_request
            )

        # no changes during this time period, so generate an EnrichedDiffs with no nodes
        if not aggregated_enriched_diffs:
            return self._build_enriched_diffs_with_no_nodes(diff_request=diff_request)

        # metadata-only diff, means that a diff exists in the database that covers at least
        # part of this time period, but it might need to have its start or end time extended
        # to cover time ranges with no changes
        if not isinstance(aggregated_enriched_diffs, EnrichedDiffs):
            return aggregated_enriched_diffs

        # a new diff (with nodes) covering the time period
        aggregated_enriched_diffs.update_metadata(
            from_time=diff_request.from_time, to_time=diff_request.to_time, tracking_id=diff_request.tracking_id
        )
        return aggregated_enriched_diffs

    async def _concatenate_diffs_and_requests(
        self,
        diff_or_request_list: Sequence[EnrichedDiffsMetadata | EnrichedDiffRequest | None],
        full_diff_request: EnrichedDiffRequest,
    ) -> EnrichedDiffs | EnrichedDiffsMetadata | None:
        """
        Returns None if diff_or_request_list is empty or all Nones
            meaning there are no changes for the diff during this time period
        Returns EnrichedDiffsMetadata if diff_or_request_list includes one EnrichedDiffsMetadata and no EnrichedDiffRequests
            meaning no diffs needed to be hydrated and combined
        Otherwise, returns EnrichedDiffs
            meaning multiple diffs (some that may have been freshly calculated) were combined
        """
        previous_diff_pair: EnrichedDiffs | EnrichedDiffsMetadata | None = None
        updated_node_uuids: set[str] = set()
        for diff_or_request in diff_or_request_list:
            if isinstance(diff_or_request, EnrichedDiffRequest):
                if previous_diff_pair:
                    log.info(f"Getting node field specifiers diff uuid={previous_diff_pair.diff_branch_diff.uuid}")
                    node_field_specifiers = await self.diff_repo.get_node_field_specifiers(
                        diff_id=previous_diff_pair.diff_branch_diff.uuid,
                    )
                    log.info(f"Number node field specifiers: {len(node_field_specifiers)}")
                    diff_or_request.node_field_specifiers = node_field_specifiers
                is_incremental_diff = diff_or_request.from_time != full_diff_request.from_time
                calculated_diff = await self._calculate_enriched_diff(
                    diff_request=diff_or_request, is_incremental_diff=is_incremental_diff
                )
                updated_node_uuids |= calculated_diff.base_node_uuids
                updated_node_uuids |= calculated_diff.branch_node_uuids
                single_enriched_diffs: EnrichedDiffs | EnrichedDiffsMetadata = calculated_diff

            elif isinstance(diff_or_request, EnrichedDiffsMetadata):
                single_enriched_diffs = diff_or_request
            else:
                continue

            if previous_diff_pair is None:
                previous_diff_pair = single_enriched_diffs
                continue

            log.info("Combining diffs...")
            previous_diff_pair = await self._combine_diffs(
                earlier=previous_diff_pair,
                later=single_enriched_diffs,
                node_uuids=updated_node_uuids,
            )
            log.info("Diffs combined.")

        return previous_diff_pair

    async def _combine_diffs(
        self,
        earlier: EnrichedDiffs | EnrichedDiffsMetadata,
        later: EnrichedDiffs | EnrichedDiffsMetadata,
        node_uuids: set[str],
    ) -> EnrichedDiffs | EnrichedDiffsMetadata:
        log.info(f"Earlier diff to combine: {earlier!r}")
        log.info(f"Later diff to combine: {later!r}")
        # if one of the diffs is hydrated and has no data, we can combine them without hydrating the other
        if isinstance(earlier, EnrichedDiffs) and earlier.is_empty:
            later.base_branch_diff.from_time = earlier.base_branch_diff.from_time
            later.diff_branch_diff.from_time = earlier.diff_branch_diff.from_time
            return later
        if isinstance(later, EnrichedDiffs) and later.is_empty:
            earlier.base_branch_diff.to_time = later.base_branch_diff.to_time
            earlier.diff_branch_diff.to_time = later.diff_branch_diff.to_time
            return earlier

        # hydrate the diffs to combine, if necessary
        if not isinstance(earlier, EnrichedDiffs):
            log.info("Hydrating earlier diff...")
            earlier = await self.diff_repo.hydrate_diff_pair(enriched_diffs_metadata=earlier, node_uuids=node_uuids)
            log.info("Earlier diff hydrated.")
        if not isinstance(later, EnrichedDiffs):
            log.info("Hydrating later diff...")
            later = await self.diff_repo.hydrate_diff_pair(enriched_diffs_metadata=later, node_uuids=node_uuids)
            log.info("Later diff hydrated.")

        return await self.diff_combiner.combine(earlier_diffs=earlier, later_diffs=later)

    async def _update_core_data_checks(self, enriched_diff: EnrichedDiffRoot | EnrichedDiffRootMetadata) -> list[Node]:
        return await self.data_check_synchronizer.synchronize(enriched_diff=enriched_diff)

    async def _calculate_enriched_diff(
        self, diff_request: EnrichedDiffRequest, is_incremental_diff: bool
    ) -> EnrichedDiffs:
        log.info(f"Calculating diff for {diff_request!r}, include_unchanged={is_incremental_diff}")
        calculated_diff_pair = await self.diff_calculator.calculate_diff(
            base_branch=diff_request.base_branch,
            diff_branch=diff_request.diff_branch,
            from_time=diff_request.from_time,
            to_time=diff_request.to_time,
            include_unchanged=is_incremental_diff,
            previous_node_specifiers=diff_request.node_field_specifiers,
        )
        log.info("Calculation complete. Enriching diff...")
        enriched_diff_pair = await self.diff_enricher.enrich(
            calculated_diffs=calculated_diff_pair, tracking_id=diff_request.tracking_id
        )
        log.info("Enrichment complete")
        return enriched_diff_pair



from infrahub import lock
from infrahub.core.registry import registry
from infrahub.message_bus import messages
from infrahub.services import InfrahubServices
from infrahub.tasks.registry import refresh_branches
from infrahub.worker import WORKER_IDENTITY


async def branches(message: messages.RefreshRegistryBranches, service: InfrahubServices) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        service.log.info("Ignoring refresh registry refresh request originating from self", worker=WORKER_IDENTITY)
        return

    async with service.database.start_session() as db:
        await refresh_branches(db=db)

    await service.component.refresh_schema_hash()


async def rebased_branch(message: messages.RefreshRegistryRebasedBranch, service: InfrahubServices) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        service.log.info(
            "Ignoring refresh registry refreshed branch for request originating from self", worker=WORKER_IDENTITY
        )
        return

    async with lock.registry.local_schema_lock():
        service.log.info("Refreshing rebased branch")
        registry.branch[message.branch] = await registry.branch_object.get_by_name(
            name=message.branch, db=service.database
        )



import os
import uuid
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional

from testcontainers.compose import DockerCompose
from typing_extensions import Self

from infrahub_testcontainers import __version__ as infrahub_version


@dataclass
class ContainerService:
    container: str
    port: int


INFRAHUB_SERVICES: dict[str, ContainerService] = {
    "server": ContainerService(container="infrahub-server-lb", port=8000),
    "task-manager": ContainerService(container="task-manager", port=4200),
}

PROJECT_ENV_VARIABLES: dict[str, str] = {
    "INFRAHUB_TESTING_DOCKER_IMAGE": "registry.opsmill.io/opsmill/infrahub",
    "INFRAHUB_TESTING_IMAGE_VERSION": infrahub_version,
    "INFRAHUB_TESTING_PRODUCTION": "false",
    "INFRAHUB_TESTING_DB_ADDRESS": "database",
    "INFRAHUB_TESTING_LOG_LEVEL": "DEBUG",
    "INFRAHUB_TESTING_GIT_REPOSITORIES_DIRECTORY": "/opt/infrahub/git",
    "INFRAHUB_TESTING_API_TOKEN": "44af444d-3b26-410d-9546-b758657e026c",
    "INFRAHUB_TESTING_INITIAL_ADMIN_TOKEN": "06438eb2-8019-4776-878c-0941b1f1d1ec",
    "INFRAHUB_TESTING_INITIAL_AGENT_TOKEN": "44af444d-3b26-410d-9546-b758657e026c",
    "INFRAHUB_TESTING_SECURITY_SECRET_KEY": "327f747f-efac-42be-9e73-999f08f86b92",
    "INFRAHUB_TESTING_ADDRESS": "http://infrahub-server-lb:8000",
    "INFRAHUB_TESTING_INTERNAL_ADDRESS": "http://infrahub-server-lb:8000",
    "INFRAHUB_TESTING_BROKER_ADDRESS": "message-queue",
    "INFRAHUB_TESTING_CACHE_ADDRESS": "cache",
    "INFRAHUB_TESTING_WORKFLOW_ADDRESS": "task-manager",
    "INFRAHUB_TESTING_TIMEOUT": "60",
    "INFRAHUB_TESTING_PREFECT_API": "http://task-manager:4200/api",
    "INFRAHUB_TESTING_LOCAL_REMOTE_GIT_DIRECTORY": "repos",
    "INFRAHUB_TESTING_INTERNAL_REMOTE_GIT_DIRECTORY": "/remote",
    "INFRAHUB_TESTING_WEB_CONCURRENCY": "4",
    "INFRAHUB_TESTING_LOCAL_DB_BACKUP_DIRECTORY": "backups",
    "INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY": "/backups",
    "INFRAHUB_TESTING_API_SERVER_COUNT": "2",
    "INFRAHUB_TESTING_TASK_WORKER_COUNT": "2",
}


@dataclass
class InfrahubDockerCompose(DockerCompose):
    project_name: Optional[str] = None

    @classmethod
    def init(cls, directory: Optional[Path] = None, version: Optional[str] = None) -> Self:
        if not directory:
            directory = Path.cwd()

        if not version:
            version = infrahub_version

        infrahub_image_version = os.environ.get("INFRAHUB_TESTING_IMAGE_VER", None)
        if version == "local" and infrahub_image_version:
            version = infrahub_image_version

        cls.create_docker_file(directory=directory)
        cls.create_env_file(directory=directory, version=version)

        return cls(project_name=cls.generate_project_name(), context=directory)

    @classmethod
    def generate_project_name(cls) -> str:
        project_id = str(uuid.uuid4())[:8]
        return f"infrahub-test-{project_id}"

    @classmethod
    def create_docker_file(cls, directory: Path) -> Path:
        current_directory = Path(__file__).resolve().parent
        compose_file = current_directory / "docker-compose.test.yml"

        test_compose_file = directory / "docker-compose.yml"
        test_compose_file.write_bytes(compose_file.read_bytes())

        haproxy_config_file = current_directory / "haproxy.cfg"

        test_haproxy_config_file = directory / "haproxy.cfg"
        test_haproxy_config_file.write_bytes(haproxy_config_file.read_bytes())

        return test_compose_file

    @classmethod
    def create_env_file(cls, directory: Path, version: str) -> Path:
        env_file = directory / ".env"

        PROJECT_ENV_VARIABLES.update({"INFRAHUB_TESTING_IMAGE_VERSION": version})

        with env_file.open(mode="w", encoding="utf-8") as file:
            for key, value in PROJECT_ENV_VARIABLES.items():
                env_var_value = os.environ.get(key, value)
                file.write(f"{key}={env_var_value}\n")
        return env_file.absolute()

    # TODO would be good to the support for project_name upstream
    @cached_property
    def compose_command_property(self) -> list[str]:
        docker_compose_cmd = [self.docker_command_path or "docker", "compose"]
        if self.compose_file_name:
            for file in self.compose_file_name:
                docker_compose_cmd += ["-f", file]
        if self.project_name:
            docker_compose_cmd += ["--project-name", self.project_name]
        if self.env_file:
            docker_compose_cmd += ["--env-file", self.env_file]
        return docker_compose_cmd

    def get_services_port(self) -> dict[str, int]:
        return {
            service_name: int(self.get_service_port(service_name=service_data.container, port=service_data.port) or 0)
            for service_name, service_data in INFRAHUB_SERVICES.items()
        }



from prefect import flow

from infrahub import config
from infrahub.core.timestamp import Timestamp
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.types import KVTTL, MessageTTL
from infrahub.services import InfrahubServices

log = get_logger()


@flow(name="validator-finalize-execution")
async def execution(message: messages.FinalizeValidatorExecution, service: InfrahubServices) -> None:
    """Monitors the status of checks associated with a validator and finalizes the conclusion of the validator

    Based on the unique execution_id this function looks expects to see an entry in the cache for each check
    associated with this validator. Upon seeing the result of a check the function will exclude it from further
    checks and update the current conclusion of the validator if any of the checks failed.

    The message will get rescheduled until the timeout has exceeded or until all checks are accounted for.
    """
    validator = await service.client.get(kind=message.validator_type, id=message.validator_id)
    checks_key = f"validator_execution_id:{message.validator_execution_id}:checks"
    current_conclusion = validator.conclusion.value
    if validator.state.value != "in_progress":
        validator.state.value = "in_progress"
        validator.started_at.value = Timestamp().to_string()
        validator.completed_at.value = ""
        await validator.save()

    required_checks_data = await service.cache.get(key=checks_key) or ""
    # Remove instances of empty checks
    required_checks = [required_check for required_check in required_checks_data.split(",") if required_check]

    completed_checks_data = await service.cache.list_keys(
        filter_pattern=f"validator_execution_id:{message.validator_execution_id}:check_execution_id:*"
    )
    completed_checks = [check.split(":")[-1] for check in completed_checks_data]

    missing_checks = [check for check in required_checks if check not in completed_checks]
    checks_to_verify = [check for check in completed_checks if check in required_checks]
    failed_check = False

    for check in checks_to_verify:
        conclusion = await service.cache.get(
            f"validator_execution_id:{message.validator_execution_id}:check_execution_id:{check}"
        )
        if conclusion != "success":
            failed_check = True

    conclusion = "failure" if failed_check else "success"
    if failed_check and current_conclusion != "failure":
        validator.conclusion.value = "failure"
        await validator.save()

    if missing_checks:
        remaining_checks = ",".join(missing_checks)
        await service.cache.set(key=checks_key, value=remaining_checks, expires=KVTTL.TWO_HOURS)
        current_time = Timestamp()
        starting_time = Timestamp(message.start_time)
        deadline = starting_time.add_delta(seconds=config.SETTINGS.miscellaneous.maximum_validator_execution_time)
        if current_time < deadline:
            log.debug(
                "Still waiting for checks to complete",
                missing_checks=missing_checks,
                validator_id=message.validator_id,
                validator_execution_id=message.validator_execution_id,
            )
            await service.send(message=message, delay=MessageTTL.FIVE)
            return

        log.info(
            "Timeout reached",
            validator_id=message.validator_id,
            validator_execution_id=message.validator_execution_id,
        )
        conclusion = "failure"

    validator.state.value = "completed"
    validator.completed_at.value = Timestamp().to_string()
    validator.conclusion.value = conclusion
    await validator.save()



from infrahub.core.node.constraints.grouped_uniqueness import NodeGroupedUniquenessConstraint
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class NodeGroupedUniquenessConstraintDependency(DependencyBuilder[NodeGroupedUniquenessConstraint]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> NodeGroupedUniquenessConstraint:
        return NodeGroupedUniquenessConstraint(db=context.db, branch=context.branch)



import importlib
from typing import Any, Awaitable, Callable, TypeVar
from uuid import UUID

from prefect.client.orchestration import PrefectClient
from prefect.client.schemas.actions import DeploymentScheduleCreate
from prefect.client.schemas.objects import FlowRun
from prefect.client.schemas.schedules import CronSchedule
from pydantic import BaseModel, Field
from typing_extensions import Self

from infrahub import __version__

from .constants import TAG_NAMESPACE, WorkflowTag, WorkflowType

TASK_RESULT_STORAGE_NAME = "infrahub-storage"

WorkflowReturn = TypeVar("WorkflowReturn")


class WorkerPoolDefinition(BaseModel):
    name: str
    worker_type: str | None = None
    description: str = ""


class WorkflowInfo(BaseModel):
    id: UUID
    info: FlowRun | None = None

    @classmethod
    def from_flow(cls, flow_run: FlowRun) -> Self:
        return cls(id=flow_run.id, info=flow_run)


class WorkflowDefinition(BaseModel):
    name: str
    type: WorkflowType = WorkflowType.INTERNAL
    module: str
    function: str
    cron: str | None = None
    tags: list[WorkflowTag] = Field(default_factory=list)

    @property
    def entrypoint(self) -> str:
        return f'backend/{self.module.replace(".", "/")}:{self.function}'

    @property
    def full_name(self) -> str:
        return f"{self.name}/{self.name}"

    def to_deployment(self) -> dict[str, Any]:
        payload: dict[str, Any] = {"name": self.name, "entrypoint": self.entrypoint, "tags": self.get_tags()}
        if self.type == WorkflowType.CORE:
            payload["version"] = __version__
        if self.cron:
            payload["schedules"] = [DeploymentScheduleCreate(schedule=CronSchedule(cron=self.cron))]

        return payload

    def get_tags(self) -> list[str]:
        tags: list[str] = []
        if self.type != WorkflowType.INTERNAL:
            tags.append(TAG_NAMESPACE)
        tags.append(WorkflowTag.WORKFLOWTYPE.render(identifier=self.type.value))
        tags += [tag.render() for tag in self.tags]
        return tags

    async def save(self, client: PrefectClient, work_pool: WorkerPoolDefinition) -> UUID:
        flow_id = await client.create_flow_from_name(self.name)
        data = self.to_deployment()
        data["work_pool_name"] = work_pool.name
        return await client.create_deployment(flow_id=flow_id, **data)

    def get_function(self) -> Callable[..., Awaitable[Any]]:
        module = importlib.import_module(self.module)
        return getattr(module, self.function)

    def validate_workflow(self) -> None:
        self.get_function()



from typing import TYPE_CHECKING, Any, Optional

from graphene import InputObjectType, Mutation
from graphql import GraphQLResolveInfo
from typing_extensions import Self

from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.mutations.main import InfrahubMutationMixin

from .main import InfrahubMutationOptions

if TYPE_CHECKING:
    from infrahub.graphql.initialization import GraphqlContext


class InfrahubGraphQLQueryMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(  # pylint: disable=arguments-differ
        cls, schema: NodeSchema, _meta: Optional[Any] = None, **options: dict[str, Any]
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)
        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def extract_query_info(
        cls, info: GraphQLResolveInfo, data: InputObjectType, branch: Branch
    ) -> dict[str, Any]:
        query_value = data.get("query", {}).get("value", None)
        if query_value is None:
            return {}

        query_info = {}
        analyzer = InfrahubGraphQLQueryAnalyzer(query=query_value, schema=info.schema, branch=branch)

        valid, errors = analyzer.is_valid
        if not valid:
            raise ValueError(f"Query is not valid, {str(errors)}")

        query_info["models"] = {"value": sorted(list(await analyzer.get_models_in_use(types=info.context.types)))}
        query_info["depth"] = {"value": await analyzer.calculate_depth()}
        query_info["height"] = {"value": await analyzer.calculate_height()}
        query_info["operations"] = {
            "value": sorted([operation.operation_type.value for operation in analyzer.operations])
        }
        query_info["variables"] = {"value": [variable.model_dump() for variable in analyzer.variables]}

        return query_info

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        data.update(await cls.extract_query_info(info=info, data=data, branch=context.branch))

        obj, result = await super().mutate_create(info=info, data=data, branch=branch)

        return obj, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: Optional[InfrahubDatabase] = None,
        node: Optional[Node] = None,
    ) -> tuple[Node, Self]:
        context: GraphqlContext = info.context

        data.update(await cls.extract_query_info(info=info, data=data, branch=context.branch))

        obj, result = await super().mutate_update(info=info, data=data, branch=branch)

        return obj, result



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core.constants import InfrahubKind
from infrahub.core.schema import SchemaRoot, core_models

from .constants import DEFAULT_MENU, MenuSection
from .models import MenuItemDefinition

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes


infrahub_schema = SchemaRoot(**core_models)


def _extract_node_icon(model: MainSchemaTypes) -> str:
    if not model.icon:
        return ""
    return model.icon


default_menu = [
    MenuItemDefinition(
        namespace="Builtin",
        name=DEFAULT_MENU,
        label=DEFAULT_MENU.title(),
        protected=True,
        icon="mdi:cube-outline",
        section=MenuSection.OBJECT,
        order_weight=10000,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="Tag",
                label="Tags",
                kind=InfrahubKind.TAG,
                protected=True,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.TAG)),
                section=MenuSection.OBJECT,
                order_weight=10000,
            )
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="IPAM",
        label="IPAM",
        protected=True,
        section=MenuSection.OBJECT,
        icon="mdi:ip-network",
        order_weight=9500,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="IPPrefix",
                label="IP Prefixes",
                kind=InfrahubKind.IPPREFIX,
                path="/ipam/prefixes",
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.IPPREFIX)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="IPAddress",
                label="IP Addresses",
                kind=InfrahubKind.IPPREFIX,
                path="/ipam/addresses?ipam-tab=ip-details",
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.IPADDRESS)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=2000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Namespaces",
                label="Namespaces",
                kind=InfrahubKind.IPNAMESPACE,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.IPNAMESPACE)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="ProposedChanges",
        label="Proposed Changes",
        path="/proposed-changes",
        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.PROPOSEDCHANGE)),
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=1000,
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="ObjectManagement",
        label="Object Management",
        icon="mdi:cube-outline",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=1500,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="Groups",
                label="Groups",
                kind=InfrahubKind.GENERICGROUP,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.GENERICGROUP)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Profiles",
                label="Profiles",
                kind=InfrahubKind.PROFILE,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.PROFILE)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=2000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="ResourceManager",
                label="Resource Manager",
                path="/resource-manager",
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.RESOURCEPOOL)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="ChangeControl",
        label="Change Control",
        icon="mdi:source-branch",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=2000,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="Branches",
                label="Branches",
                path="/branches",
                icon="mdi:layers-triple",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="CheckDefinition",
                label="Check Definition",
                kind=InfrahubKind.CHECKDEFINITION,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.CHECKDEFINITION)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Tasks",
                label="Tasks",
                path="/tasks",
                icon="mdi:shield-check",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="UnifiedStorage",
        label="Unified Storage",
        icon="mdi:nas",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=2500,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="Schema",
                label="Schema",
                path="/schema",
                icon="mdi:file-code",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Git Repository",
                label="Repository",
                kind=InfrahubKind.GENERICREPOSITORY,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.GENERICREPOSITORY)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=2000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Credentials",
                label="Credentials",
                kind=InfrahubKind.CREDENTIAL,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.CREDENTIAL)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=2000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="GraphqlQuery",
                label="GraphQL Query",
                kind=InfrahubKind.GRAPHQLQUERY,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.GRAPHQLQUERY)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="Deployment",
        label="Deployment",
        icon="mdi:rocket-launch",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=3000,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="ArtifactMenu",
                label="Artifact",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
                children=[
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="Artifact",
                        label="Artifact",
                        kind=InfrahubKind.ARTIFACT,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.ARTIFACT)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=1000,
                    ),
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="ArtifactDefinition",
                        label="Artifact Definition",
                        kind=InfrahubKind.ARTIFACTDEFINITION,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.ARTIFACTDEFINITION)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=1000,
                    ),
                ],
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="GeneratorMenu",
                label="Generator",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
                children=[
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="GeneratorInstance",
                        label="Generator Instance",
                        kind=InfrahubKind.GENERATORINSTANCE,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.GENERATORINSTANCE)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=1000,
                    ),
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="GeneratorDefinition",
                        label="Generator Definition",
                        kind=InfrahubKind.GENERATORDEFINITION,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.GENERATORDEFINITION)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=2000,
                    ),
                ],
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Transformation",
                label="Transformation",
                kind=InfrahubKind.TRANSFORM,
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.TRANSFORM)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="Integration",
        label="Integrations",
        icon="mdi:connection",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=3500,
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="Webhooks",
                label="Webhooks",
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.CUSTOMWEBHOOK)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=3000,
                children=[
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="WebhookStandard",
                        label="Webhook",
                        kind=InfrahubKind.STANDARDWEBHOOK,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.STANDARDWEBHOOK)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=1000,
                    ),
                    MenuItemDefinition(
                        namespace="Builtin",
                        name="WebhookCustom",
                        label="Custom Webhook",
                        kind=InfrahubKind.CUSTOMWEBHOOK,
                        icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.CUSTOMWEBHOOK)),
                        protected=True,
                        section=MenuSection.INTERNAL,
                        order_weight=2000,
                    ),
                ],
            ),
        ],
    ),
    MenuItemDefinition(
        namespace="Builtin",
        name="Admin",
        label="Admin",
        icon="mdi:settings-outline",
        protected=True,
        section=MenuSection.INTERNAL,
        order_weight=10000,
        permissions=["global:super_admin:allow_all"],
        children=[
            MenuItemDefinition(
                namespace="Builtin",
                name="RoleManagement",
                label="Users & Permissions",
                path="/role-management",
                icon=_extract_node_icon(infrahub_schema.get(InfrahubKind.BASEPERMISSION)),
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=1000,
            ),
            MenuItemDefinition(
                namespace="Builtin",
                name="Menu",
                label="Menu",
                kind=InfrahubKind.MENU,
                icon="mdi:menu",
                protected=True,
                section=MenuSection.INTERNAL,
                order_weight=2500,
            ),
        ],
    ),
]



from __future__ import annotations

from pathlib import Path
from typing import Optional

from pydantic import BaseModel

from infrahub.exceptions import Error
from infrahub.git.constants import BRANCHES_DIRECTORY_NAME, COMMITS_DIRECTORY_NAME
from infrahub.git.directory import get_repositories_directory


class Worktree(BaseModel):
    identifier: str
    directory: Path
    commit: str
    branch: Optional[str] = None

    @classmethod
    def init(cls, text: str) -> Worktree:
        lines = text.split("\n")

        full_directory = Path(lines[0].replace("worktree ", ""))

        # Extract the identifier from the directory name
        # We first need to substract the main repository_directory to get the relative path.
        repo_directory = get_repositories_directory()
        relative_directory = full_directory.relative_to(repo_directory)
        relative_paths = relative_directory.parts

        identifier = None
        if len(relative_paths) == 2:
            # this is the main worktree for the main branch
            identifier = relative_paths[1]
        elif len(relative_paths) == 3 and relative_paths[1] == COMMITS_DIRECTORY_NAME:
            # this is the either a commit or a branch worktree
            identifier = relative_paths[2]
        elif len(relative_paths) == 3 and relative_paths[1] == BRANCHES_DIRECTORY_NAME and lines[2] != "detached":
            identifier = lines[2].replace("branch refs/heads/", "")
        else:
            raise Error("Unexpected path for a worktree.")

        item = cls(identifier=identifier, directory=full_directory, commit=lines[1].replace("HEAD ", ""))

        if lines[2] != "detached":
            item.branch = lines[2].replace("branch refs/heads/", "")

        return item



import importlib.metadata

try:
    __version__ = importlib.metadata.version("infrahub-testcontainers")
except importlib.metadata.PackageNotFoundError:
    __version__ = importlib.metadata.version("infrahub-server")



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from infrahub.message_bus.types import KVTTL
    from infrahub.services import InfrahubServices


class InfrahubCache:
    """Base class for caching services"""

    async def initialize(self, service: InfrahubServices) -> None:
        """Initialize the cache"""

    async def delete(self, key: str) -> None:
        """Delete a key from the cache."""
        raise NotImplementedError()

    async def get(self, key: str) -> Optional[str]:
        """Retrieve a value from the cache."""
        raise NotImplementedError()

    async def get_values(self, keys: list[str]) -> list[Optional[str]]:
        """Return a list the values for requested keys."""
        raise NotImplementedError()

    async def list_keys(self, filter_pattern: str) -> list[str]:
        """Return a list of active keys that match the provided filter."""
        raise NotImplementedError()

    async def set(
        self, key: str, value: str, expires: Optional[KVTTL] = None, not_exists: bool = False
    ) -> Optional[bool]:
        """Set a value in the cache."""
        raise NotImplementedError()



from collections import defaultdict
from enum import Enum
from typing import Iterable, Optional, Union

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import RelationshipDeleteBehavior
from infrahub.core.node import Node
from infrahub.core.query.relationship import (
    FullRelationshipIdentifier,
    RelationshipGetByIdentifierQuery,
    RelationshipPeersData,
)
from infrahub.core.schema import MainSchemaTypes, NodeSchema, ProfileSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import ValidationError


class DeleteRelationshipType(Enum):
    CASCADE_DELETE = "cascade"
    DEPENDENT_NODE = "dependent"


class NodeDeleteIndex:
    def __init__(self, all_schemas_map: dict[str, MainSchemaTypes]) -> None:
        self._all_schemas_map = all_schemas_map
        # {node_schema: {DeleteRelationshipType: {relationship_identifier: peer_node_schema}}}
        self._dependency_graph: dict[str, dict[DeleteRelationshipType, dict[str, set[str]]]] = {}

    def index(self, start_schemas: Iterable[NodeSchema | ProfileSchema]) -> None:
        self._index_cascading_deletes(start_schemas=start_schemas)
        self._index_dependent_schema(start_schemas=start_schemas)

    def _get_schema_kinds(self, schema_kind: str) -> set[str]:
        schema = self._all_schemas_map[schema_kind]
        if schema.is_node_schema:
            return {schema_kind}
        used_by = getattr(schema, "used_by", None)
        if schema.is_generic_schema and used_by:
            return set(used_by)
        return set()

    def _add_to_dependency_graph(
        self, kind: str, relationship_type: DeleteRelationshipType, relationship_identifier: str, peer_kinds: set[str]
    ) -> None:
        if kind not in self._dependency_graph:
            self._dependency_graph[kind] = {}
        if relationship_type not in self._dependency_graph[kind]:
            self._dependency_graph[kind][relationship_type] = defaultdict(set)
        self._dependency_graph[kind][relationship_type][relationship_identifier].update(peer_kinds)

    def _index_cascading_deletes(self, start_schemas: Iterable[NodeSchema | ProfileSchema]) -> None:
        kinds_to_check: set[str] = {schema.kind for schema in start_schemas}
        while True:
            try:
                kind_to_check = kinds_to_check.pop()
            except KeyError:
                break
            node_schema = self._all_schemas_map[kind_to_check]
            for relationship_schema in node_schema.relationships:
                if relationship_schema.on_delete != RelationshipDeleteBehavior.CASCADE:
                    continue
                peer_kinds = self._get_schema_kinds(schema_kind=relationship_schema.peer)
                self._add_to_dependency_graph(
                    kind=kind_to_check,
                    relationship_type=DeleteRelationshipType.CASCADE_DELETE,
                    relationship_identifier=relationship_schema.get_identifier(),
                    peer_kinds=peer_kinds,
                )
                for peer_kind in peer_kinds:
                    if peer_kind not in self._dependency_graph:
                        kinds_to_check.add(peer_kind)

    def _index_dependent_schema(self, start_schemas: Iterable[NodeSchema | ProfileSchema]) -> None:
        start_schema_kinds: set[str] = set()
        for start_schema in start_schemas:
            start_schema_kinds.add(start_schema.kind)
            if start_schema.inherit_from:
                start_schema_kinds.update(set(start_schema.inherit_from))
        for node_schema in self._all_schemas_map.values():
            for relationship_schema in node_schema.relationships:
                if relationship_schema.optional is True or relationship_schema.peer not in start_schema_kinds:
                    continue

                for peer_kind in self._get_schema_kinds(schema_kind=relationship_schema.peer):
                    self._add_to_dependency_graph(
                        kind=peer_kind,
                        relationship_type=DeleteRelationshipType.DEPENDENT_NODE,
                        relationship_identifier=relationship_schema.get_identifier(),
                        peer_kinds={node_schema.kind},
                    )

    def get_relationship_identifiers(self) -> list[FullRelationshipIdentifier]:
        full_relationship_identifiers = []
        for node_kind, relationship_type_details in self._dependency_graph.items():
            for relationship_map in relationship_type_details.values():
                for relationship_identifier, peer_kinds in relationship_map.items():
                    full_relationship_identifiers.extend(
                        [
                            FullRelationshipIdentifier(
                                source_kind=node_kind, identifier=relationship_identifier, destination_kind=peer_kind
                            )
                            for peer_kind in peer_kinds
                        ]
                    )
        return full_relationship_identifiers

    def get_relationship_types(self, src_kind: str, relationship_identifier: str) -> set[DeleteRelationshipType]:
        relationship_types: set[DeleteRelationshipType] = set()
        if src_kind not in self._dependency_graph:
            return relationship_types
        for relationship_type, relationships_map in self._dependency_graph[src_kind].items():
            if relationship_identifier in relationships_map:
                relationship_types.add(relationship_type)
        return relationship_types


class NodeDeleteValidator:
    def __init__(self, db: InfrahubDatabase, branch: Branch):
        self.db = db
        self.branch = branch
        schema_branch = registry.schema.get_schema_branch(name=self.branch.name)
        self._all_schemas_map = schema_branch.get_all(duplicate=False)
        self.index: NodeDeleteIndex = NodeDeleteIndex(all_schemas_map=self._all_schemas_map)

    async def get_ids_to_delete(self, nodes: Iterable[Node], at: Optional[Union[Timestamp, str]] = None) -> set[str]:
        start_schemas = {node.get_schema() for node in nodes}
        self.index.index(start_schemas=start_schemas)
        at = Timestamp(at)

        return await self._analyze_delete_dependencies(start_nodes=nodes, at=at)

    async def _analyze_delete_dependencies(
        self, start_nodes: Iterable[Node], at: Optional[Union[Timestamp, str]]
    ) -> set[str]:
        full_relationship_identifiers = self.index.get_relationship_identifiers()
        if not full_relationship_identifiers:
            return {node.get_id() for node in start_nodes}

        query = await RelationshipGetByIdentifierQuery.init(
            db=self.db, full_identifiers=full_relationship_identifiers, branch=self.branch, at=at
        )
        await query.execute(db=self.db)

        peer_data_by_source_id = self._build_peer_data_map(peers_datas=query.get_peers())
        node_ids_to_check = {node.get_id() for node in start_nodes}
        node_ids_to_delete: set[str] = set()
        dependent_node_details_map: dict[str, list[RelationshipPeersData]] = {}

        while node_ids_to_check:
            node_id = node_ids_to_check.pop()
            node_ids_to_delete.add(node_id)
            if node_id not in peer_data_by_source_id:
                continue
            peer_data_list = peer_data_by_source_id[node_id]
            for peer_data in peer_data_list:
                relationship_types = self.index.get_relationship_types(
                    src_kind=peer_data.source_kind, relationship_identifier=peer_data.identifier
                )
                peer_id = str(peer_data.destination_id)
                if DeleteRelationshipType.CASCADE_DELETE in relationship_types:
                    if peer_id not in node_ids_to_delete:
                        node_ids_to_check.add(peer_id)
                if DeleteRelationshipType.DEPENDENT_NODE in relationship_types:
                    if peer_id not in dependent_node_details_map:
                        dependent_node_details_map[peer_id] = []
                    dependent_node_details_map[peer_id].append(peer_data)

        missing_delete_ids = set(dependent_node_details_map.keys()) - node_ids_to_delete
        if not missing_delete_ids:
            return node_ids_to_delete
        missing_delete_peers_data = []
        for peers_data_list in dependent_node_details_map.values():
            missing_delete_peers_data.extend(peers_data_list)
        validation_error = self._build_validation_error(missing_delete_peers_data=missing_delete_peers_data)
        raise validation_error

    def _build_peer_data_map(
        self, peers_datas: Iterable[RelationshipPeersData]
    ) -> dict[str, list[RelationshipPeersData]]:
        peer_data_by_source_id: dict[str, list[RelationshipPeersData]] = {}
        for peer_data in peers_datas:
            source_id = str(peer_data.source_id)
            if source_id not in peer_data_by_source_id:
                peer_data_by_source_id[source_id] = []
            peer_data_by_source_id[source_id].append(peer_data)
            # check if this relationship also needs to be tracked going the other way
            if not self.index.get_relationship_types(
                src_kind=peer_data.destination_kind, relationship_identifier=peer_data.identifier
            ):
                continue
            dest_id = str(peer_data.destination_id)
            if dest_id not in peer_data_by_source_id:
                peer_data_by_source_id[dest_id] = []
            peer_data_by_source_id[dest_id].append(peer_data.reversed())
        return peer_data_by_source_id

    def _build_validation_error(self, missing_delete_peers_data: Iterable[RelationshipPeersData]) -> ValidationError:
        validation_errors = []
        for peers_data in missing_delete_peers_data:
            peer_kind = peers_data.destination_kind
            peer_schema = self._all_schemas_map[peer_kind]
            peer_rel_name = peer_schema.get_relationship_by_identifier(peers_data.identifier).name
            peer_path = f"{peer_kind}.{peer_rel_name}"
            err_msg = f"Cannot delete {peers_data.source_kind} '{peers_data.source_id}'."
            err_msg += f" It is linked to mandatory relationship {peer_rel_name} on node {peer_kind} '{peers_data.destination_id}'"
            validation_errors.append(ValidationError({peer_path: err_msg}))

        return ValidationError(validation_errors)



import os
import shutil
import subprocess  # noqa: S404
from pathlib import Path

import pytest

from infrahub_testcontainers import __version__ as infrahub_version

from .container import PROJECT_ENV_VARIABLES, InfrahubDockerCompose


class TestInfrahubDocker:
    @pytest.fixture(scope="class")
    def infrahub_version(self) -> str:
        return infrahub_version

    def execute_ctl_run(self, address: str, script: str) -> str:
        env = os.environ.copy()
        env["INFRAHUB_ADDRESS"] = address
        env["INFRAHUB_API_TOKEN"] = PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_INITIAL_ADMIN_TOKEN"]
        env["INFRAHUB_MAX_CONCURRENT_EXECUTION"] = "1"
        result = subprocess.run(  # noqa: S602
            f"infrahubctl run {script}", shell=True, capture_output=True, text=True, env=env, check=False
        )
        return result.stdout

    @pytest.fixture(scope="class")
    def tmp_directory(self, tmpdir_factory: pytest.TempdirFactory) -> Path:
        directory = Path(str(tmpdir_factory.getbasetemp().strpath))
        return directory

    @pytest.fixture(scope="class")
    def remote_repos_dir(self, tmp_directory: Path) -> Path:
        directory = tmp_directory / PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_LOCAL_REMOTE_GIT_DIRECTORY"]
        directory.mkdir(exist_ok=True)

        return directory

    @pytest.fixture(scope="class")
    def remote_backups_dir(self, tmp_directory: Path) -> Path:
        directory = tmp_directory / PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_LOCAL_DB_BACKUP_DIRECTORY"]
        directory.mkdir(exist_ok=True)

        return directory

    @pytest.fixture(scope="class")
    def default_branch(self) -> str:
        return "main"

    @pytest.fixture(scope="class")
    def infrahub_compose(
        self,
        tmp_directory: Path,
        remote_repos_dir: Path,  # initialize repository before running docker compose to fix permissions issues # noqa: ARG002
        remote_backups_dir: Path,  # noqa: ARG002
        infrahub_version: str,
    ) -> InfrahubDockerCompose:
        return InfrahubDockerCompose.init(directory=tmp_directory, version=infrahub_version)

    @pytest.fixture(scope="class")
    def infrahub_app(self, request: pytest.FixtureRequest, infrahub_compose: InfrahubDockerCompose) -> dict[str, int]:
        def cleanup() -> None:
            infrahub_compose.stop()

        request.addfinalizer(cleanup)

        infrahub_compose.start()

        return infrahub_compose.get_services_port()

    @pytest.fixture(scope="class")
    def infrahub_port(self, infrahub_app: dict[str, int]) -> int:
        return infrahub_app["server"]

    @pytest.fixture(scope="class")
    def task_manager_port(self, infrahub_app: dict[str, int]) -> int:
        return infrahub_app["task-manager"]

    def backup_database(self, request: pytest.FixtureRequest, dest_dir: Path | None = None) -> None:
        assert "enterprise" in os.environ.get("NEO4J_DOCKER_IMAGE", "")

        backup_dir: Path = request.getfixturevalue("remote_backups_dir")
        infrahub_compose: InfrahubDockerCompose = request.getfixturevalue("infrahub_compose")

        infrahub_compose.exec_in_container(
            command=[
                "neo4j-admin",
                "database",
                "backup",
                "--to-path",
                os.environ.get(
                    "INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY",
                    PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY"],
                ),
            ],
            service_name="database",
        )

        if dest_dir:
            shutil.copytree(
                str(backup_dir),
                str(dest_dir),
            )

    def restore_database(self, request: pytest.FixtureRequest, backup_file: Path) -> None:
        assert "enterprise" in os.environ.get("NEO4J_DOCKER_IMAGE", "")

        backup_dir: Path = request.getfixturevalue("remote_backups_dir")
        infrahub_compose: InfrahubDockerCompose = request.getfixturevalue("infrahub_compose")

        shutil.copy(
            str(backup_file),
            str(backup_dir / backup_file.name),
        )

        infrahub_compose.exec_in_container(
            command=["cypher-shell", "-u", "neo4j", "-p", "admin", "STOP DATABASE neo4j;"],
            service_name="database",
        )

        infrahub_compose.exec_in_container(
            command=[
                "neo4j-admin",
                "database",
                "restore",
                "--overwrite-destination",
                "--from-path",
                str(
                    Path(
                        os.environ.get(
                            "INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY",
                            PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY"],
                        )
                    )
                    / backup_file.name
                ),
            ],
            service_name="database",
        )

        infrahub_compose.exec_in_container(
            command=["cypher-shell", "-d", "system", "-u", "neo4j", "-p", "admin", "START DATABASE neo4j;"],
            service_name="database",
        )

        infrahub_compose.stop(down=False)
        infrahub_compose.start()



from prefect import flow

from infrahub.exceptions import RepositoryError
from infrahub.git.repository import InfrahubRepository, get_initialized_repo, initialize_repo
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.messages.git_repository_connectivity import (
    GitRepositoryConnectivityResponse,
    GitRepositoryConnectivityResponseData,
)
from infrahub.services import InfrahubServices
from infrahub.worker import WORKER_IDENTITY

log = get_logger()


@flow(name="git-repository-check-connectivity", flow_run_name="Check connectivity for {message.repository_name}")
async def connectivity(message: messages.GitRepositoryConnectivity, service: InfrahubServices) -> None:
    response_data = GitRepositoryConnectivityResponseData(message="Successfully accessed repository", success=True)

    try:
        InfrahubRepository.check_connectivity(name=message.repository_name, url=message.repository_location)
    except RepositoryError as exc:
        response_data.success = False
        response_data.message = exc.message

    if message.reply_requested:
        response = GitRepositoryConnectivityResponse(
            data=response_data,
        )
        await service.reply(message=response, initiator=message)


@flow(name="refresh-git-fetch", flow_run_name="Fetch git repository {message.repository_name} on " + WORKER_IDENTITY)
async def fetch(message: messages.RefreshGitFetch, service: InfrahubServices) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        log.info("Ignoring git fetch request originating from self", worker=WORKER_IDENTITY)
        return

    try:
        repo = await get_initialized_repo(
            repository_id=message.repository_id,
            name=message.repository_name,
            service=service,
            repository_kind=message.repository_kind,
        )
    except RepositoryError:
        repo = await initialize_repo(
            location=message.location,
            repository_id=message.repository_id,
            name=message.repository_name,
            service=service,
            repository_kind=message.repository_kind,
        )

    await repo.fetch()
    await repo.pull(
        branch_name=message.infrahub_branch_name,
        branch_id=message.infrahub_branch_id,
        create_if_missing=True,
        update_commit_value=False,
    )



from __future__ import annotations

from enum import IntFlag

from prefect import flow, task
from prefect.logging import get_run_logger
from pydantic import BaseModel

from infrahub import lock
from infrahub.core.constants import CheckType, InfrahubKind, RepositoryInternalStatus
from infrahub.core.diff.coordinator import DiffCoordinator
from infrahub.core.registry import registry
from infrahub.dependencies.registry import get_component_registry
from infrahub.git.repository import InfrahubRepository
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.message_bus.types import (
    ProposedChangeArtifactDefinition,
    ProposedChangeBranchDiff,
    ProposedChangeRepository,
    ProposedChangeSubscriber,
)
from infrahub.proposed_change.models import (
    RequestProposedChangeDataIntegrity,
    RequestProposedChangeRepositoryChecks,
    RequestProposedChangeRunGenerators,
    RequestProposedChangeSchemaIntegrity,
    RequestProposedChangeUserTests,
)
from infrahub.services import InfrahubServices  # noqa: TC001
from infrahub.workflows.catalogue import (
    REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS,
    REQUEST_PROPOSED_CHANGE_RUN_GENERATORS,
    REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
    REQUEST_PROPOSED_CHANGE_USER_TESTS,
)
from infrahub.workflows.utils import add_tags


class DefinitionSelect(IntFlag):
    NONE = 0
    MODIFIED_KINDS = 1
    FILE_CHANGES = 2

    @staticmethod
    def add_flag(current: DefinitionSelect, flag: DefinitionSelect, condition: bool) -> DefinitionSelect:
        if condition:
            return current | flag
        return current

    @property
    def log_line(self) -> str:
        change_types = []
        if DefinitionSelect.MODIFIED_KINDS in self:
            change_types.append("data changes within relevant object kinds")

        if DefinitionSelect.FILE_CHANGES in self:
            change_types.append("file modifications in Git repositories")

        if self:
            return f"Requesting generation due to {' and '.join(change_types)}"

        return "Doesn't require changes due to no relevant modified kinds or file changes in Git"


@flow(name="proposed-changed-pipeline", flow_run_name="Execute Pipeline")
async def pipeline(message: messages.RequestProposedChangePipeline, service: InfrahubServices) -> None:
    events: list[InfrahubMessage] = []

    repositories = await _get_proposed_change_repositories(message=message, service=service)

    if message.source_branch_sync_with_git and await _validate_repository_merge_conflicts(
        repositories=repositories, service=service
    ):
        for repo in repositories:
            if not repo.read_only and repo.internal_status == RepositoryInternalStatus.ACTIVE.value:
                events.append(
                    messages.RequestRepositoryChecks(
                        proposed_change=message.proposed_change,
                        repository=repo.repository_id,
                        source_branch=repo.source_branch,
                        target_branch=repo.destination_branch,
                    )
                )
        for event in events:
            event.assign_meta(parent=message)
            await service.send(message=event)
        return

    await _gather_repository_repository_diffs(repositories=repositories, service=service)

    async with service.database.start_transaction() as dbt:
        destination_branch = await registry.get_branch(db=dbt, branch=message.destination_branch)
        source_branch = await registry.get_branch(db=dbt, branch=message.source_branch)
        component_registry = get_component_registry()
        diff_coordinator = await component_registry.get_component(DiffCoordinator, db=dbt, branch=source_branch)
        await diff_coordinator.update_branch_diff(base_branch=destination_branch, diff_branch=source_branch)

    diff_summary = await service.client.get_diff_summary(branch=message.source_branch)
    branch_diff = ProposedChangeBranchDiff(diff_summary=diff_summary, repositories=repositories)
    await _populate_subscribers(branch_diff=branch_diff, service=service, branch=message.source_branch)

    if message.check_type is CheckType.ARTIFACT:
        events.append(
            messages.RequestProposedChangeRefreshArtifacts(
                proposed_change=message.proposed_change,
                source_branch=message.source_branch,
                source_branch_sync_with_git=message.source_branch_sync_with_git,
                destination_branch=message.destination_branch,
                branch_diff=branch_diff,
            )
        )

    if message.check_type in [CheckType.ALL, CheckType.GENERATOR]:
        model_proposed_change_run_generator = RequestProposedChangeRunGenerators(
            proposed_change=message.proposed_change,
            source_branch=message.source_branch,
            source_branch_sync_with_git=message.source_branch_sync_with_git,
            destination_branch=message.destination_branch,
            branch_diff=branch_diff,
            refresh_artifacts=message.check_type is CheckType.ALL,
            do_repository_checks=message.check_type is CheckType.ALL,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_RUN_GENERATORS, parameters={"model": model_proposed_change_run_generator}
        )

    if message.check_type in [CheckType.ALL, CheckType.DATA] and branch_diff.has_node_changes(
        branch=message.source_branch
    ):
        model_proposed_change_data_integrity = RequestProposedChangeDataIntegrity(
            proposed_change=message.proposed_change,
            source_branch=message.source_branch,
            source_branch_sync_with_git=message.source_branch_sync_with_git,
            destination_branch=message.destination_branch,
            branch_diff=branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_DATA_INTEGRITY, parameters={"model": model_proposed_change_data_integrity}
        )

    if message.check_type in [CheckType.REPOSITORY, CheckType.USER]:
        model_proposed_change_repo_checks = RequestProposedChangeRepositoryChecks(
            proposed_change=message.proposed_change,
            source_branch=message.source_branch,
            source_branch_sync_with_git=message.source_branch_sync_with_git,
            destination_branch=message.destination_branch,
            branch_diff=branch_diff,
        )
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_REPOSITORY_CHECKS, parameters={"model": model_proposed_change_repo_checks}
        )

    if message.check_type in [CheckType.ALL, CheckType.SCHEMA] and branch_diff.has_data_changes(
        branch=message.source_branch
    ):
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_SCHEMA_INTEGRITY,
            parameters={
                "model": RequestProposedChangeSchemaIntegrity(
                    proposed_change=message.proposed_change,
                    source_branch=message.source_branch,
                    source_branch_sync_with_git=message.source_branch_sync_with_git,
                    destination_branch=message.destination_branch,
                    branch_diff=branch_diff,
                )
            },
        )

    if message.check_type in [CheckType.ALL, CheckType.TEST]:
        await service.workflow.submit_workflow(
            workflow=REQUEST_PROPOSED_CHANGE_USER_TESTS,
            parameters={
                "model": RequestProposedChangeUserTests(
                    proposed_change=message.proposed_change,
                    source_branch=message.source_branch,
                    source_branch_sync_with_git=message.source_branch_sync_with_git,
                    destination_branch=message.destination_branch,
                    branch_diff=branch_diff,
                )
            },
        )

    for event in events:
        event.assign_meta(parent=message)
        await service.send(message=event)


@flow(
    name="proposed-changed-refresh-artifact",
    flow_run_name="Trigger artifacts refresh",
)
async def refresh_artifacts(message: messages.RequestProposedChangeRefreshArtifacts, service: InfrahubServices) -> None:
    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change])
    log = get_run_logger()

    definition_information = await service.client.execute_graphql(
        query=GATHER_ARTIFACT_DEFINITIONS,
        branch_name=message.source_branch,
    )
    artifact_definitions = _parse_artifact_definitions(
        definitions=definition_information[InfrahubKind.ARTIFACTDEFINITION]["edges"]
    )

    for artifact_definition in artifact_definitions:
        # Request artifact definition checks if the source branch that is managed in combination
        # to the Git repository containing modifications which could indicate changes to the transforms
        # in code
        # Alternatively if the queries used touches models that have been modified in the path
        # impacted artifact definitions will be included for consideration

        select = DefinitionSelect.NONE
        select = select.add_flag(
            current=select,
            flag=DefinitionSelect.FILE_CHANGES,
            condition=message.source_branch_sync_with_git and message.branch_diff.has_file_modifications,
        )

        for changed_model in message.branch_diff.modified_kinds(branch=message.source_branch):
            condition = False
            if (changed_model in artifact_definition.query_models) or (
                changed_model.startswith("Profile")
                and changed_model.replace("Profile", "", 1) in artifact_definition.query_models
            ):
                condition = True

            select = select.add_flag(
                current=select,
                flag=DefinitionSelect.MODIFIED_KINDS,
                condition=condition,
            )

        if select:
            log.info(f"Trigger processing of {artifact_definition.definition_name}")
            msg = messages.RequestArtifactDefinitionCheck(
                artifact_definition=artifact_definition,
                branch_diff=message.branch_diff,
                proposed_change=message.proposed_change,
                source_branch=message.source_branch,
                source_branch_sync_with_git=message.source_branch_sync_with_git,
                destination_branch=message.destination_branch,
            )

            msg.assign_meta(parent=message)
            await service.send(message=msg)


GATHER_ARTIFACT_DEFINITIONS = """
query GatherArtifactDefinitions {
  CoreArtifactDefinition {
    edges {
      node {
        id
        name {
          value
        }
        artifact_name {
          value
        }
        content_type {
            value
        }
        transformation {
          node {
            __typename
            timeout {
                value
            }
            query {
              node {
                models {
                  value
                }
                name {
                  value
                }
              }
            }
            ... on CoreTransformJinja2 {
              template_path {
                value
              }
            }
            ... on CoreTransformPython {
              class_name {
                value
              }
              file_path {
                value
              }
            }
            repository {
              node {
                id
              }
            }
          }
        }
      }
    }
  }
}
"""

GATHER_GRAPHQL_QUERY_SUBSCRIBERS = """
query GatherGraphQLQuerySubscribers($members: [ID!]) {
  CoreGraphQLQueryGroup(members__ids: $members) {
    edges {
      node {
        subscribers {
          edges {
            node {
              id
              __typename
            }
          }
        }
      }
    }
  }
}
"""


DESTINATION_ALLREPOSITORIES = """
query DestinationBranchRepositories {
  CoreGenericRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        ... on CoreRepository {
          commit {
            value
          }
        }
        ... on CoreReadOnlyRepository {
          commit {
            value
          }
        }
      }
    }
  }
}
"""

SOURCE_REPOSITORIES = """
query MyQuery {
  CoreRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        commit {
          value
        }
      }
    }
  }
}
"""
SOURCE_READONLY_REPOSITORIES = """
query MyQuery {
  CoreReadOnlyRepository {
    edges {
      node {
        __typename
        id
        name {
          value
        }
        internal_status {
          value
        }
        commit {
          value
        }
      }
    }
  }
}
"""


class Repository(BaseModel):
    repository_id: str
    repository_name: str
    read_only: bool
    commit: str
    internal_status: str


def _parse_proposed_change_repositories(
    message: messages.RequestProposedChangePipeline, source: list[dict], destination: list[dict]
) -> list[ProposedChangeRepository]:
    """This function assumes that the repos is a list of the edges

    The data should come from the queries:
    * DESTINATION_ALLREPOSITORIES
    * SOURCE_REPOSITORIES
    * SOURCE_READONLY_REPOSITORIES
    """
    destination_repos = _parse_repositories(repositories=destination)
    source_repos = _parse_repositories(repositories=source)
    pc_repos: dict[str, ProposedChangeRepository] = {}
    for repo in destination_repos:
        if repo.repository_id not in pc_repos:
            pc_repos[repo.repository_id] = ProposedChangeRepository(
                repository_id=repo.repository_id,
                repository_name=repo.repository_name,
                read_only=repo.read_only,
                internal_status=repo.internal_status,
                destination_commit=repo.commit,
                source_branch=message.source_branch,
                destination_branch=message.destination_branch,
            )
        else:
            pc_repos[repo.repository_id].destination_commit = repo.commit

    for repo in source_repos:
        if repo.repository_id not in pc_repos:
            pc_repos[repo.repository_id] = ProposedChangeRepository(
                repository_id=repo.repository_id,
                repository_name=repo.repository_name,
                read_only=repo.read_only,
                internal_status=repo.internal_status,
                source_commit=repo.commit,
                source_branch=message.source_branch,
                destination_branch=message.destination_branch,
            )
        else:
            pc_repos[repo.repository_id].source_commit = repo.commit
            pc_repos[repo.repository_id].internal_status = repo.internal_status

    return list(pc_repos.values())


def _parse_repositories(repositories: list[dict]) -> list[Repository]:
    """This function assumes that the repos is a list of the edges

    The data should come from the queries:
    * DESTINATION_ALLREPOSITORIES
    * SOURCE_REPOSITORIES
    * SOURCE_READONLY_REPOSITORIES
    """
    parsed = []
    for repo in repositories:
        parsed.append(
            Repository(
                repository_id=repo["node"]["id"],
                repository_name=repo["node"]["name"]["value"],
                read_only=repo["node"]["__typename"] == InfrahubKind.READONLYREPOSITORY,
                commit=repo["node"]["commit"]["value"] or "",
                internal_status=repo["node"]["internal_status"]["value"],
            )
        )
    return parsed


def _parse_artifact_definitions(definitions: list[dict]) -> list[ProposedChangeArtifactDefinition]:
    """This function assumes that definitions is a list of the edges

    The edge should be of type CoreArtifactDefinition from the query
    * GATHER_ARTIFACT_DEFINITIONS
    """

    parsed = []
    for definition in definitions:
        artifact_definition = ProposedChangeArtifactDefinition(
            definition_id=definition["node"]["id"],
            definition_name=definition["node"]["name"]["value"],
            artifact_name=definition["node"]["artifact_name"]["value"],
            content_type=definition["node"]["content_type"]["value"],
            timeout=definition["node"]["transformation"]["node"]["timeout"]["value"],
            query_name=definition["node"]["transformation"]["node"]["query"]["node"]["name"]["value"],
            query_models=definition["node"]["transformation"]["node"]["query"]["node"]["models"]["value"] or [],
            repository_id=definition["node"]["transformation"]["node"]["repository"]["node"]["id"],
            transform_kind=definition["node"]["transformation"]["node"]["__typename"],
        )
        if artifact_definition.transform_kind == InfrahubKind.TRANSFORMJINJA2:
            artifact_definition.template_path = definition["node"]["transformation"]["node"]["template_path"]["value"]
        elif artifact_definition.transform_kind == InfrahubKind.TRANSFORMPYTHON:
            artifact_definition.class_name = definition["node"]["transformation"]["node"]["class_name"]["value"]
            artifact_definition.file_path = definition["node"]["transformation"]["node"]["file_path"]["value"]

        parsed.append(artifact_definition)

    return parsed


async def _get_proposed_change_repositories(
    message: messages.RequestProposedChangePipeline, service: InfrahubServices
) -> list[ProposedChangeRepository]:
    destination_all = await service.client.execute_graphql(
        query=DESTINATION_ALLREPOSITORIES, branch_name=message.destination_branch
    )
    source_managed = await service.client.execute_graphql(query=SOURCE_REPOSITORIES, branch_name=message.source_branch)
    source_readonly = await service.client.execute_graphql(
        query=SOURCE_READONLY_REPOSITORIES, branch_name=message.source_branch
    )

    destination_all = destination_all[InfrahubKind.GENERICREPOSITORY]["edges"]
    source_all = (
        source_managed[InfrahubKind.REPOSITORY]["edges"] + source_readonly[InfrahubKind.READONLYREPOSITORY]["edges"]
    )

    return _parse_proposed_change_repositories(message=message, source=source_all, destination=destination_all)


@task(name="proposed-change-validate-repository-conflicts", task_run_name="Validate conflicts on repository")
async def _validate_repository_merge_conflicts(
    repositories: list[ProposedChangeRepository], service: InfrahubServices
) -> bool:
    log = get_run_logger()
    conflicts = False
    for repo in repositories:
        if repo.has_diff and not repo.is_staging:
            git_repo = await InfrahubRepository.init(
                id=repo.repository_id, name=repo.repository_name, client=service.client
            )
            async with lock.registry.get(name=repo.repository_name, namespace="repository"):
                repo.conflicts = await git_repo.get_conflicts(
                    source_branch=repo.source_branch, dest_branch=repo.destination_branch
                )
                if repo.conflicts:
                    log.info(f"{len(repo.conflicts)} conflict(s) identified on {repo.repository_name}")
                    conflicts = True
                else:
                    log.info(f"no conflict identified for {repo.repository_name}")

    return conflicts


async def _gather_repository_repository_diffs(
    repositories: list[ProposedChangeRepository], service: InfrahubServices
) -> None:
    for repo in repositories:
        if repo.has_diff and repo.source_commit and repo.destination_commit:
            # TODO we need to find a way to return all files in the repo if the repo is new
            git_repo = await InfrahubRepository.init(
                id=repo.repository_id, name=repo.repository_name, client=service.client
            )

            files_changed: list[str] = []
            files_added: list[str] = []
            files_removed: list[str] = []

            if repo.destination_branch:
                files_changed, files_added, files_removed = await git_repo.calculate_diff_between_commits(
                    first_commit=repo.source_commit, second_commit=repo.destination_commit
                )
            else:
                files_added = await git_repo.list_all_files(commit=repo.source_commit)

            repo.files_removed = files_removed
            repo.files_added = files_added
            repo.files_changed = files_changed


async def _populate_subscribers(branch_diff: ProposedChangeBranchDiff, service: InfrahubServices, branch: str) -> None:
    result = await service.client.execute_graphql(
        query=GATHER_GRAPHQL_QUERY_SUBSCRIBERS,
        branch_name=branch,
        variables={"members": branch_diff.modified_nodes(branch=branch)},
    )

    for group in result[InfrahubKind.GRAPHQLQUERYGROUP]["edges"]:
        for subscriber in group["node"]["subscribers"]["edges"]:
            branch_diff.subscribers.append(
                ProposedChangeSubscriber(subscriber_id=subscriber["node"]["id"], kind=subscriber["node"]["__typename"])
            )



import ipaddress
from enum import Enum
from typing import Union

IPNetworkType = Union[ipaddress.IPv6Network, ipaddress.IPv4Network]
IPAddressType = Union[ipaddress.IPv6Interface, ipaddress.IPv4Interface]
AllIPTypes = Union[IPNetworkType, IPAddressType]


class PrefixMemberType(Enum):
    PREFIX = "prefix"
    ADDRESS = "address"



from infrahub_sdk.protocols import CoreCheckDefinition
from infrahub_sdk.uuidt import UUIDT
from prefect import flow
from prefect.logging import get_run_logger

from infrahub import lock
from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import CheckError
from infrahub.git.repository import InfrahubRepository
from infrahub.message_bus import InfrahubMessage, messages
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices
from infrahub.workflows.utils import add_tags


@flow(
    name="git-repository-check-definition",
    flow_run_name="Run user defined checks for repository {message.repository_name}",
)
async def check_definition(message: messages.CheckRepositoryCheckDefinition, service: InfrahubServices) -> None:
    await add_tags(branches=[message.branch_name], nodes=[message.proposed_change])
    log = get_run_logger()

    definition = await service.client.get(
        kind=CoreCheckDefinition, id=message.check_definition_id, branch=message.branch_name
    )
    proposed_change = await service.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=message.proposed_change)
    validator_execution_id = str(UUIDT())
    check_execution_ids: list[str] = []
    await proposed_change.validations.fetch()
    validator = None
    events: list[InfrahubMessage] = []

    for relationship in proposed_change.validations.peers:
        existing_validator = relationship.peer

        if (
            existing_validator.typename == InfrahubKind.USERVALIDATOR
            and existing_validator.repository.id == message.repository_id
            and existing_validator.check_definition.id == message.check_definition_id
        ):
            validator = existing_validator
            service.log.info("Found the same validator", validator=validator)

    if validator:
        validator.conclusion.value = "unknown"
        validator.state.value = "queued"
        validator.started_at.value = ""
        validator.completed_at.value = ""
        await validator.save()
    else:
        validator = await service.client.create(
            kind=InfrahubKind.USERVALIDATOR,
            data={
                "label": f"Check: {definition.name.value}",
                "proposed_change": message.proposed_change,
                "repository": message.repository_id,
                "check_definition": message.check_definition_id,
            },
        )
        await validator.save()

    if definition.targets.id:
        # Check against a group of targets
        await definition.targets.fetch()
        group = definition.targets.peer
        await group.members.fetch()
        for relationship in group.members.peers:
            member = relationship.peer

            check_execution_id = str(UUIDT())
            check_execution_ids.append(check_execution_id)
            events.append(
                messages.CheckRepositoryUserCheck(
                    name=member.display_label,
                    validator_id=validator.id,
                    validator_execution_id=validator_execution_id,
                    check_execution_id=check_execution_id,
                    repository_id=message.repository_id,
                    repository_name=message.repository_name,
                    commit=message.commit,
                    file_path=message.file_path,
                    class_name=message.class_name,
                    branch_name=message.branch_name,
                    check_definition_id=message.check_definition_id,
                    proposed_change=message.proposed_change,
                    variables=member.extract(params=definition.parameters.value),
                    branch_diff=message.branch_diff,
                    timeout=definition.timeout.value,
                )
            )

    else:
        check_execution_id = str(UUIDT())
        check_execution_ids.append(check_execution_id)
        events.append(
            messages.CheckRepositoryUserCheck(
                name=definition.name.value,
                validator_id=validator.id,
                validator_execution_id=validator_execution_id,
                check_execution_id=check_execution_id,
                repository_id=message.repository_id,
                repository_name=message.repository_name,
                commit=message.commit,
                file_path=message.file_path,
                class_name=message.class_name,
                branch_name=message.branch_name,
                check_definition_id=message.check_definition_id,
                proposed_change=message.proposed_change,
                branch_diff=message.branch_diff,
                timeout=definition.timeout.value,
            )
        )

    checks_in_execution = ",".join(check_execution_ids)
    log.info(f"Checks in execution {checks_in_execution}")
    await service.cache.set(
        key=f"validator_execution_id:{validator_execution_id}:checks",
        value=checks_in_execution,
        expires=KVTTL.TWO_HOURS,
    )
    events.append(
        messages.FinalizeValidatorExecution(
            start_time=Timestamp().to_string(),
            validator_id=validator.id,
            validator_execution_id=validator_execution_id,
            validator_type=InfrahubKind.USERVALIDATOR,
        )
    )

    for event in events:
        event.assign_meta(parent=message)
        await service.send(message=event)


@flow(
    name="git-repository-check-merge-conflict",
    flow_run_name="Check for merge conflicts between {message.source_branch} and {message.target_branch}",
)
async def merge_conflicts(message: messages.CheckRepositoryMergeConflicts, service: InfrahubServices) -> None:
    """Runs a check to see if there are merge conflicts between two branches."""
    await add_tags(branches=[message.source_branch], nodes=[message.proposed_change])

    success_condition = "-"
    validator = await service.client.get(kind=InfrahubKind.REPOSITORYVALIDATOR, id=message.validator_id)
    await validator.checks.fetch()

    repo = await InfrahubRepository.init(id=message.repository_id, name=message.repository_name, service=service)
    async with lock.registry.get(name=message.repository_name, namespace="repository"):
        conflicts = await repo.get_conflicts(source_branch=message.source_branch, dest_branch=message.target_branch)

    existing_checks = {}
    for relationship in validator.checks.peers:
        existing_check = relationship.peer
        if existing_check.typename == InfrahubKind.FILECHECK and existing_check.kind.value == "MergeConflictCheck":
            check_key = ""
            if existing_check.files.value:
                check_key = "".join(existing_check.files.value)
            check_key = f"-{check_key}"
            existing_checks[check_key] = existing_check

    validator_conclusion = "success"
    check = None
    if conflicts:
        validator_conclusion = "failure"
        for conflict in conflicts:
            conflict_key = f"-{conflict}"
            if conflict_key in existing_checks:
                existing_checks[conflict_key].created_at.value = Timestamp().to_string()
                await existing_checks[conflict_key].save()
                existing_checks.pop(conflict_key)
            else:
                check = await service.client.create(
                    kind=InfrahubKind.FILECHECK,
                    data={
                        "name": conflict,
                        "origin": "ConflictCheck",
                        "kind": "MergeConflictCheck",
                        "validator": message.validator_id,
                        "created_at": Timestamp().to_string(),
                        "files": [conflict],
                        "conclusion": "failure",
                        "severity": "critical",
                    },
                )
                await check.save()

    elif success_condition in existing_checks:
        existing_checks[success_condition].created_at.value = Timestamp().to_string()
        await existing_checks[success_condition].save()
        existing_checks.pop(success_condition)

    else:
        check = await service.client.create(
            kind=InfrahubKind.FILECHECK,
            data={
                "name": "Merge Conflict Check",
                "origin": "ConflictCheck",
                "kind": "MergeConflictCheck",
                "validator": message.validator_id,
                "created_at": Timestamp().to_string(),
                "conclusion": "success",
                "severity": "info",
            },
        )
        await check.save()

    async with service.database.start_session() as db:
        await NodeManager.delete(db=db, nodes=list(existing_checks.values()))

    await service.cache.set(
        key=f"validator_execution_id:{message.validator_execution_id}:check_execution_id:{message.check_execution_id}",
        value=validator_conclusion,
        expires=KVTTL.TWO_HOURS,
    )


@flow(name="git-repository-user-check", flow_run_name="Execute user defined Check '{message.name}'")
async def user_check(message: messages.CheckRepositoryUserCheck, service: InfrahubServices) -> None:
    await add_tags(branches=[message.branch_name], nodes=[message.proposed_change])
    log = get_run_logger()

    validator = await service.client.get(kind=InfrahubKind.USERVALIDATOR, id=message.validator_id)
    await validator.checks.fetch()

    repo = await InfrahubRepository.init(
        id=message.repository_id, name=message.repository_name, commit=message.commit, service=service
    )
    conclusion = "failure"
    severity = "critical"
    log_entries = ""
    try:
        check_run = await repo.execute_python_check.with_options(timeout_seconds=message.timeout)(
            branch_name=message.branch_name,
            location=message.file_path,
            class_name=message.class_name,
            client=service.client,
            commit=message.commit,
            params=message.variables,
        )
        if check_run.passed:
            conclusion = "success"
            severity = "info"
            log.info("The check passed")
        else:
            log.warning("The check reported failures")
            for log_entry in check_run.log_entries:
                log.warning(log_entry)
        log_entries = check_run.log_entries
    except CheckError as exc:
        log.warning("The check failed to run")
        log.error(exc.message)
        log_entries = f"FATAL Error/n:{exc.message}"

    check = None
    for relationship in validator.checks.peers:
        existing_check = relationship.peer
        if (
            existing_check.typename == InfrahubKind.STANDARDCHECK
            and existing_check.kind.value == "CheckDefinition"
            and existing_check.name.value == message.name
        ):
            check = existing_check

    if check:
        check.created_at.value = Timestamp().to_string()
        check.message.value = log_entries
        check.conclusion.value = conclusion
        check.severity.value = severity
        await check.save()
    else:
        check = await service.client.create(
            kind=InfrahubKind.STANDARDCHECK,
            data={
                "name": message.name,
                "origin": message.repository_id,
                "kind": "CheckDefinition",
                "validator": message.validator_id,
                "created_at": Timestamp().to_string(),
                "message": log_entries,
                "conclusion": conclusion,
                "severity": severity,
            },
        )
        await check.save()

    await service.cache.set(
        key=f"validator_execution_id:{message.validator_execution_id}:check_execution_id:{message.check_execution_id}",
        value=conclusion,
        expires=KVTTL.TWO_HOURS,
    )



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, Union

from pydantic import BaseModel, ConfigDict, Field, field_validator

from .._importer import import_module
from ..checks import InfrahubCheck
from ..exceptions import (
    ModuleImportError,
    ResourceNotDefinedError,
)
from ..generator import InfrahubGenerator
from ..transforms import InfrahubTransform
from ..utils import duplicates

if TYPE_CHECKING:
    from ..node import InfrahubNode, InfrahubNodeSync

    InfrahubNodeTypes = Union[InfrahubNode, InfrahubNodeSync]

ResourceClass = TypeVar("ResourceClass")


class InfrahubRepositoryConfigElement(BaseModel):
    """Class to regroup all elements of the infrahub configuration for a repository for typing purpose."""


class InfrahubRepositoryArtifactDefinitionConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the artifact definition")
    artifact_name: str | None = Field(default=None, description="Name of the artifact created from this definition")
    parameters: dict[str, Any] = Field(..., description="The input parameters required to render this artifact")
    content_type: str = Field(..., description="The content type of the rendered artifact")
    targets: str = Field(..., description="The group to target when creating artifacts")
    transformation: str = Field(..., description="The transformation to use.")


class InfrahubJinja2TransformConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the transform")
    query: str = Field(..., description="The name of the GraphQL Query")
    template_path: Path = Field(..., description="The path within the repository of the template file")
    description: str | None = Field(default=None, description="Description for this transform")

    @property
    def template_path_value(self) -> str:
        return str(self.template_path)

    @property
    def payload(self) -> dict[str, str]:
        data = self.model_dump(exclude_none=True)
        data["template_path"] = self.template_path_value
        return data


class InfrahubCheckDefinitionConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the Check Definition")
    file_path: Path = Field(..., description="The file within the repository with the check code.")
    parameters: dict[str, Any] = Field(
        default_factory=dict, description="The input parameters required to run this check"
    )
    targets: str | None = Field(
        default=None, description="The group to target when running this check, leave blank for global checks"
    )
    class_name: str = Field(default="Check", description="The name of the check class to run.")

    def load_class(self, import_root: str | None = None, relative_path: str | None = None) -> type[InfrahubCheck]:
        module = import_module(module_path=self.file_path, import_root=import_root, relative_path=relative_path)

        if self.class_name not in dir(module):
            raise ModuleImportError(message=f"The specified class {self.class_name} was not found within the module")

        check_class = getattr(module, self.class_name)

        if not issubclass(check_class, InfrahubCheck):
            raise ModuleImportError(message=f"The specified class {self.class_name} is not an Infrahub Check")

        return check_class


class InfrahubGeneratorDefinitionConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the Generator Definition")
    file_path: Path = Field(..., description="The file within the repository with the generator code.")
    query: str = Field(..., description="The GraphQL query to use as input.")
    parameters: dict[str, Any] = Field(
        default_factory=dict, description="The input parameters required to run this check"
    )
    targets: str = Field(..., description="The group to target when running this generator")
    class_name: str = Field(default="Generator", description="The name of the generator class to run.")
    convert_query_response: bool = Field(
        default=False,
        description="Decide if the generator should convert the result of the GraphQL query to SDK InfrahubNode objects.",
    )

    def load_class(self, import_root: str | None = None, relative_path: str | None = None) -> type[InfrahubGenerator]:
        module = import_module(module_path=self.file_path, import_root=import_root, relative_path=relative_path)

        if self.class_name not in dir(module):
            raise ModuleImportError(message=f"The specified class {self.class_name} was not found within the module")

        generator_class = getattr(module, self.class_name)

        if not issubclass(generator_class, InfrahubGenerator):
            raise ModuleImportError(message=f"The specified class {self.class_name} is not an Infrahub Generator")

        return generator_class


class InfrahubPythonTransformConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the Transform")
    file_path: Path = Field(..., description="The file within the repository with the transform code.")
    class_name: str = Field(default="Transform", description="The name of the transform class to run.")

    def load_class(self, import_root: str | None = None, relative_path: str | None = None) -> type[InfrahubTransform]:
        module = import_module(module_path=self.file_path, import_root=import_root, relative_path=relative_path)

        if self.class_name not in dir(module):
            raise ModuleImportError(message=f"The specified class {self.class_name} was not found within the module")

        transform_class = getattr(module, self.class_name)

        if not issubclass(transform_class, InfrahubTransform):
            raise ModuleImportError(message=f"The specified class {self.class_name} is not an Infrahub Transform")

        return transform_class


class InfrahubRepositoryGraphQLConfig(InfrahubRepositoryConfigElement):
    model_config = ConfigDict(extra="forbid")
    name: str = Field(..., description="The name of the GraphQL Query")
    file_path: Path = Field(..., description="The file within the repository with the query code.")

    def load_query(self, relative_path: str = ".") -> str:
        file_name = Path(f"{relative_path}/{self.file_path}")
        with file_name.open("r", encoding="UTF-8") as file:
            return file.read()


RESOURCE_MAP: dict[Any, str] = {
    InfrahubJinja2TransformConfig: "jinja2_transforms",
    InfrahubCheckDefinitionConfig: "check_definitions",
    InfrahubRepositoryArtifactDefinitionConfig: "artifact_definitions",
    InfrahubPythonTransformConfig: "python_transforms",
    InfrahubGeneratorDefinitionConfig: "generator_definitions",
    InfrahubRepositoryGraphQLConfig: "queries",
}


class InfrahubRepositoryConfig(BaseModel):
    model_config = ConfigDict(extra="forbid")
    check_definitions: list[InfrahubCheckDefinitionConfig] = Field(
        default_factory=list, description="User defined checks"
    )
    schemas: list[Path] = Field(default_factory=list, description="Schema files")
    jinja2_transforms: list[InfrahubJinja2TransformConfig] = Field(
        default_factory=list, description="Jinja2 data transformations"
    )
    artifact_definitions: list[InfrahubRepositoryArtifactDefinitionConfig] = Field(
        default_factory=list, description="Artifact definitions"
    )
    python_transforms: list[InfrahubPythonTransformConfig] = Field(
        default_factory=list, description="Python data transformations"
    )
    generator_definitions: list[InfrahubGeneratorDefinitionConfig] = Field(
        default_factory=list, description="Generator definitions"
    )
    queries: list[InfrahubRepositoryGraphQLConfig] = Field(default_factory=list, description="GraphQL Queries")

    @field_validator(
        "check_definitions",
        "jinja2_transforms",
        "artifact_definitions",
        "python_transforms",
        "generator_definitions",
        "queries",
    )
    @classmethod
    def unique_items(cls, v: list[Any]) -> list[Any]:
        names = [item.name for item in v]
        if dups := duplicates(names):
            raise ValueError(f"Found multiples element with the same names: {dups}")
        return v

    def _has_resource(self, resource_id: str, resource_type: type[ResourceClass], resource_field: str = "name") -> bool:
        for item in getattr(self, RESOURCE_MAP[resource_type]):
            if getattr(item, resource_field) == resource_id:
                return True
        return False

    def _get_resource(
        self, resource_id: str, resource_type: type[ResourceClass], resource_field: str = "name"
    ) -> ResourceClass:
        for item in getattr(self, RESOURCE_MAP[resource_type]):
            if getattr(item, resource_field) == resource_id:
                return item
        raise ResourceNotDefinedError(f"Unable to find {resource_id!r} in {RESOURCE_MAP[resource_type]!r}")

    def has_jinja2_transform(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubJinja2TransformConfig)

    def get_jinja2_transform(self, name: str) -> InfrahubJinja2TransformConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubJinja2TransformConfig)

    def has_check_definition(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubCheckDefinitionConfig)

    def get_check_definition(self, name: str) -> InfrahubCheckDefinitionConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubCheckDefinitionConfig)

    def has_artifact_definition(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubRepositoryArtifactDefinitionConfig)

    def get_artifact_definition(self, name: str) -> InfrahubRepositoryArtifactDefinitionConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubRepositoryArtifactDefinitionConfig)

    def has_generator_definition(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubGeneratorDefinitionConfig)

    def get_generator_definition(self, name: str) -> InfrahubGeneratorDefinitionConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubGeneratorDefinitionConfig)

    def has_python_transform(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubPythonTransformConfig)

    def get_python_transform(self, name: str) -> InfrahubPythonTransformConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubPythonTransformConfig)

    def has_query(self, name: str) -> bool:
        return self._has_resource(resource_id=name, resource_type=InfrahubRepositoryGraphQLConfig)

    def get_query(self, name: str) -> InfrahubRepositoryGraphQLConfig:
        return self._get_resource(resource_id=name, resource_type=InfrahubRepositoryGraphQLConfig)



from typing import Any

from neo4j.graph import Node as Neo4jNode

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffConflictQuery(Query):
    name = "enriched_diff_conflict"
    type = QueryType.READ

    def __init__(self, conflict_id: str, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.conflict_id = conflict_id

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:
        self.params = {"conflict_id": self.conflict_id}
        query = "MATCH (conflict:DiffConflict {uuid: $conflict_id})"
        self.return_labels = ["conflict"]
        self.add_to_query(query=query)

    async def get_conflict_node(self) -> Neo4jNode | None:
        result = self.get_result()
        if not result:
            return None
        return result.get_node("conflict")



from typing import Optional

from infrahub_sdk.client import Config as InfrahubClientConfig
from infrahub_sdk.client import InfrahubClientSync
from infrahub_sdk.node import InfrahubNodeSync
from pytest import Config, Item, Session, TestReport

from infrahub.core.constants import InfrahubKind
from infrahub.core.timestamp import Timestamp

OUTCOME_TO_CONCLUSION_MAP = {"passed": "success", "failed": "failure", "skipped": "failure"}
OUTCOME_TO_SEVERITY_MAP = {"passed": "success", "failed": "error", "skipped": "warning"}
ORDER_TYPE_MAP = {"infrahub_smoke": 0, "infrahub_unit": 1, "infrahub_integration": 2}
ORDER_ITEM_MAP = {
    "infrahub_check": 0,
    "infrahub_graphql_query": 1,
    "infrahub_jinja2_transform": 2,
    "infrahub_python_transform": 3,
}


class InfrahubBackendPlugin:
    def __init__(self, config: InfrahubClientConfig, repository_id: str, proposed_change_id: str) -> None:
        self.client = InfrahubClientSync(config=config)

        self.repository_id = repository_id
        self.proposed_change_id = proposed_change_id

        self.proposed_change: InfrahubNodeSync
        self.validator: InfrahubNodeSync
        self.checks: dict[str, InfrahubNodeSync] = {}

    def get_repository_validator(self) -> tuple[InfrahubNodeSync, bool]:
        """Return the existing RepositoryValidator for the ProposedChange or create a new one."""
        validator_name = "Repository Tests Validator"

        for relationship in self.proposed_change.validations.peers:
            existing_validator = relationship.peer

            if (
                existing_validator.typename == InfrahubKind.REPOSITORYVALIDATOR
                and existing_validator.repository.id == self.repository_id
                and existing_validator.label.value == validator_name
            ):
                return existing_validator, False

        validator = self.client.create(
            kind=InfrahubKind.REPOSITORYVALIDATOR,
            data={"label": validator_name, "proposed_change": self.proposed_change, "repository": self.repository_id},
        )
        validator.save()

        return validator, True

    def pytest_collection_modifyitems(self, session: Session, config: Config, items: list[Item]) -> None:  # pylint: disable=unused-argument
        """This function is called after item collection and gives the opportunity to work on the collection before sending the items for testing.

        All items without an "infrahub" marker will be discarded. Items will also be re-ordered to be run in a specific order:
        1. Smoke tests
        2. Unit tests
        3. Integration tests

        TODO: this function should also filter items that do not need to run
        """
        filtered_items = [i for i in items if i.get_closest_marker("infrahub")]

        def sort_key(item: Item) -> tuple[int, int]:
            type_cost = 99
            for marker_name, priority in ORDER_TYPE_MAP.items():
                if item.get_closest_marker(marker_name):
                    type_cost = priority
                    break

            item_cost = 99
            for marker_name, priority in ORDER_ITEM_MAP.items():
                if item.get_closest_marker(marker_name):
                    type_cost = priority
                    break

            return type_cost, item_cost

        filtered_items.sort(key=sort_key)
        items[:] = filtered_items

    def pytest_collection_finish(self, session: Session) -> None:  # pylint: disable=unused-argument
        """This function is called when tests have been collected and modified, meaning they are ready to be run."""
        self.proposed_change = self.client.get(kind=InfrahubKind.PROPOSEDCHANGE, id=self.proposed_change_id)
        self.proposed_change.validations.fetch()

        self.validator, is_new_validator = self.get_repository_validator()
        # Workaround for https://github.com/opsmill/infrahub/issues/2184
        if not is_new_validator:
            self.validator.checks.fetch()
            for relationship in self.validator.checks.peers:
                check = relationship.peer
                self.checks[check.origin.value] = check

    def pytest_runtestloop(self, session: Session) -> Optional[object]:  # pylint: disable=unused-argument,useless-return
        """This function is called when the test loop is being run."""
        self.validator.conclusion.value = "unknown"
        self.validator.state.value = "in_progress"
        self.validator.started_at.value = Timestamp().to_string()
        self.validator.completed_at.value = ""

        return None

    def pytest_runtest_setup(self, item: Item) -> None:
        """Create a StandardCheck for each test item to later record its details.

        If a check already exists, reset it to its default values.
        """
        created_at = Timestamp().to_string()
        check = self.checks.get(item.nodeid, None)
        if check:
            check.message.value = ""
            check.conclusion.value = "unknown"
            check.severity.value = "info"
            check.created_at.value = created_at
        else:
            check = self.client.create(
                kind=InfrahubKind.STANDARDCHECK,
                data={
                    "name": item.name,
                    "origin": item.nodeid,
                    "kind": "TestReport",
                    "validator": self.validator.id,
                    "created_at": created_at,
                    "severity": "info",
                },
            )
            self.checks[item.nodeid] = check

        check.save()

    def pytest_runtest_logreport(self, report: TestReport) -> None:
        """This function is called 3 times per test: setup, call, teardown."""
        if report.when != "call":
            return

        check = self.checks[report.nodeid]
        check.message.value = report.longreprtext
        check.severity.value = OUTCOME_TO_SEVERITY_MAP[report.outcome]
        check.conclusion.value = OUTCOME_TO_CONCLUSION_MAP[report.outcome]
        # Workaround for https://github.com/opsmill/infrahub/issues/2184
        check.update(do_full_update=True)

    def pytest_sessionfinish(self, session: Session) -> None:  # pylint: disable=unused-argument
        """Set the final RepositoryValidator details after completing the test session."""
        conclusion = "success"

        for check in self.checks.values():
            if check.conclusion.value == "failure":
                conclusion = check.conclusion.value
                break

        self.validator.state.value = "completed"
        self.validator.completed_at.value = Timestamp().to_string()
        self.validator.conclusion.value = conclusion
        self.validator.save()



from infrahub.core.validators.relationship.optional import RelationshipOptionalChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaRelationshipOptionalConstraintDependency(DependencyBuilder[RelationshipOptionalChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> RelationshipOptionalChecker:
        return RelationshipOptionalChecker(db=context.db, branch=context.branch)



from collections import defaultdict
from uuid import UUID

from prefect.client.schemas.objects import Log as PrefectLog
from pydantic import BaseModel, Field

from .constants import LOG_LEVEL_MAPPING


class RelatedNodeInfo(BaseModel):
    id: str
    kind: str | None = None


class RelatedNodesInfo(BaseModel):
    flows: dict[UUID, dict[str, RelatedNodeInfo]] = Field(default_factory=lambda: defaultdict(dict))
    nodes: dict[str, RelatedNodeInfo] = Field(default_factory=dict)

    def add_nodes(self, flow_id: UUID, node_ids: list[str]) -> None:
        for node_id in node_ids:
            self.add_node(flow_id=flow_id, node_id=node_id)

    def add_node(self, flow_id: UUID, node_id: str) -> None:
        if node_id not in self.nodes:
            node = RelatedNodeInfo(id=node_id)
            self.nodes[node_id] = node
        self.flows[flow_id][node_id] = self.nodes[node_id]

    def get_related_nodes(self, flow_id: UUID) -> list[RelatedNodeInfo]:
        if flow_id not in self.flows or len(self.flows[flow_id].keys()) == 0:
            return []
        return list(self.flows[flow_id].values())

    def get_related_nodes_as_dict(self, flow_id: UUID) -> list[dict[str, str | None]]:
        if flow_id not in self.flows or len(self.flows[flow_id].keys()) == 0:
            return []
        return [item.model_dump() for item in list(self.flows[flow_id].values())]

    def get_first_related_node(self, flow_id: UUID) -> RelatedNodeInfo | None:
        if nodes := self.get_related_nodes(flow_id=flow_id):
            return nodes[0]
        return None

    def get_unique_related_node_ids(self) -> list[str]:
        return list(self.nodes.keys())


class FlowLogs(BaseModel):
    logs: defaultdict[UUID, list[PrefectLog]] = Field(default_factory=lambda: defaultdict(list))

    def to_graphql(self, flow_id: UUID) -> list[dict]:
        return [
            {
                "node": {
                    "message": log.message,
                    "severity": LOG_LEVEL_MAPPING.get(log.level, "error"),
                    "timestamp": log.timestamp.to_iso8601_string(),
                }
            }
            for log in self.logs[flow_id]
        ]


class FlowProgress(BaseModel):
    data: dict[UUID, float] = Field(default_factory=dict)



from typing import TypedDict


class RelationshipMergeDict(TypedDict):
    peer_id: str
    name: str
    action: str


class AttributeMergeDict(TypedDict):
    name: str
    action: str


class NodeMergeDict(TypedDict):
    uuid: str
    action: str
    attributes: list[AttributeMergeDict]
    relationships: list[RelationshipMergeDict]


class PropertyMergeDict(TypedDict):
    property_type: str
    action: str
    value: str | bool | int | float | None


class AttributePropertyMergeDict(TypedDict):
    node_uuid: str
    attribute_name: str
    properties: list[PropertyMergeDict]


class RelationshipPropertyMergeDict(TypedDict):
    node_uuid: str
    relationship_id: str
    peer_uuid: str
    properties: list[PropertyMergeDict]



from dataclasses import dataclass
from typing import Optional

from infrahub.core.constants import DiffAction
from infrahub.core.constants.database import DatabaseEdgeType
from infrahub.core.diff.model.path import BranchTrackingId
from infrahub.core.ipam.kinds_getter import IpamKindsGetter
from infrahub.core.ipam.model import IpamNodeDetails
from infrahub.core.manager import NodeManager
from infrahub.database import InfrahubDatabase

from .model.path import EnrichedDiffNode
from .repository.repository import DiffRepository


@dataclass
class ChangedIpamNodeDetails:
    node_uuid: str
    is_address: bool
    is_delete: bool
    namespace_id: Optional[str]
    ip_value: Optional[str]


class IpamDiffParser:
    def __init__(
        self,
        db: InfrahubDatabase,
        diff_repository: DiffRepository,
        ip_kinds_getter: IpamKindsGetter,
    ) -> None:
        self.db = db
        self.diff_repo = diff_repository
        self.ip_kinds_getter = ip_kinds_getter

    async def get_changed_ipam_node_details(
        self, source_branch_name: str, target_branch_name: str
    ) -> list[IpamNodeDetails]:
        ip_address_kinds = await self.ip_kinds_getter.get_ipam_address_kinds(
            branch_names=[source_branch_name, target_branch_name]
        )
        ip_prefix_kinds = await self.ip_kinds_getter.get_ipam_prefix_kinds(
            branch_names=[source_branch_name, target_branch_name]
        )
        if not ip_address_kinds and not ip_prefix_kinds:
            return []

        enriched_diffs = await self.diff_repo.get(
            base_branch_name=target_branch_name,
            diff_branch_names=[source_branch_name],
            tracking_id=BranchTrackingId(name=source_branch_name),
            filters={
                "kind": {"includes": list(ip_address_kinds | ip_prefix_kinds)},
                "status": {"excludes": {DiffAction.UNCHANGED}},
            },
        )
        changed_node_details: list[ChangedIpamNodeDetails] = []
        for diff in enriched_diffs:
            for node_diff in diff.nodes:
                if node_diff.action is DiffAction.UNCHANGED:
                    continue
                if node_diff.kind in ip_address_kinds:
                    is_address = True
                elif node_diff.kind in ip_prefix_kinds:
                    is_address = False
                else:
                    continue
                ip_value = self._get_ip_value(node_diff=node_diff)
                namespace_id = self._get_namespace_id(node_diff=node_diff)
                changed_node_details.append(
                    ChangedIpamNodeDetails(
                        node_uuid=node_diff.uuid,
                        is_delete=node_diff.action is DiffAction.REMOVED,
                        is_address=is_address,
                        namespace_id=namespace_id,
                        ip_value=ip_value,
                    )
                )
        await self._add_missing_values(
            source_branch_name=source_branch_name,
            target_branch_name=target_branch_name,
            changed_node_details=changed_node_details,
        )

        return [
            IpamNodeDetails(
                node_uuid=cnd.node_uuid,
                is_delete=cnd.is_delete,
                is_address=cnd.is_address,
                namespace_id=cnd.namespace_id,
                ip_value=cnd.ip_value,
            )
            for cnd in changed_node_details
            if cnd.namespace_id and cnd.ip_value
        ]

    async def _add_missing_values_branch(
        self, branch_name: str, changed_node_details: list[ChangedIpamNodeDetails], uuids_missing_data: set[str]
    ) -> None:
        nodes = await NodeManager.get_many(
            db=self.db, branch=branch_name, ids=list(uuids_missing_data), prefetch_relationships=True
        )

        for cnd in changed_node_details:
            if cnd.ip_value and cnd.namespace_id:
                continue
            node_from_db = nodes.get(cnd.node_uuid)
            if not node_from_db:
                continue
            if not cnd.ip_value:
                if cnd.is_address and hasattr(node_from_db, "address"):
                    cnd.ip_value = node_from_db.address.value
                elif not cnd.is_address and hasattr(node_from_db, "prefix"):
                    cnd.ip_value = node_from_db.prefix.value
            if not cnd.namespace_id:
                rels = await node_from_db.ip_namespace.get_relationships(db=self.db)  # type: ignore[attr-defined]
                if rels:
                    cnd.namespace_id = rels[0].get_peer_id()
            if cnd.ip_value and cnd.namespace_id and cnd.node_uuid in uuids_missing_data:
                uuids_missing_data.remove(cnd.node_uuid)

    async def _add_missing_values(
        self, source_branch_name: str, target_branch_name: str, changed_node_details: list[ChangedIpamNodeDetails]
    ) -> None:
        uuids_missing_data = {
            cnd.node_uuid for cnd in changed_node_details if cnd.ip_value is None or cnd.namespace_id is None
        }
        if not uuids_missing_data:
            return

        await self._add_missing_values_branch(
            branch_name=source_branch_name,
            changed_node_details=changed_node_details,
            uuids_missing_data=uuids_missing_data,
        )
        if not uuids_missing_data:
            return

        await self._add_missing_values_branch(
            branch_name=target_branch_name,
            changed_node_details=changed_node_details,
            uuids_missing_data=uuids_missing_data,
        )

    def _get_ip_value(self, node_diff: EnrichedDiffNode) -> Optional[str]:
        ip_attr_diff = None
        for diff_attr in node_diff.attributes:
            if diff_attr.name in {"prefix", "address"}:
                ip_attr_diff = diff_attr
                break
        if not ip_attr_diff:
            return None
        for diff_property in ip_attr_diff.properties:
            if diff_property.property_type is DatabaseEdgeType.HAS_VALUE:
                return diff_property.new_value or diff_property.previous_value
        return None

    def _get_namespace_id(self, node_diff: EnrichedDiffNode) -> Optional[str]:
        namespace_rel = None
        for diff_rel in node_diff.relationships:
            if diff_rel.name == "ip_namespace":
                namespace_rel = diff_rel
                break
        if not namespace_rel or not namespace_rel.relationships:
            return None
        namespace_rel_element = next(iter(namespace_rel.relationships))
        return namespace_rel_element.peer_id



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, InputObjectType, Mutation, String

from infrahub.core.account import ObjectPermission
from infrahub.core.constants import ComputedAttributeKind, MutationAction, PermissionAction, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.registry import registry
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta, NodeMutatedEvent
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.log import get_log_data
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class InfrahubComputedAttributeUpdateInput(InputObjectType):
    id = String(required=True)
    kind = String(required=True)
    attribute = String(required=True)
    value = String(required=True)


class UpdateComputedAttribute(Mutation):
    class Arguments:
        data = InfrahubComputedAttributeUpdateInput(required=True)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_computed_attribute")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubComputedAttributeUpdateInput,
    ) -> UpdateComputedAttribute:
        context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(name=str(data.kind), branch=context.branch.name, duplicate=False)
        target_attribute = node_schema.get_attribute(name=str(data.attribute))
        if (
            not target_attribute.computed_attribute
            or target_attribute.computed_attribute.kind == ComputedAttributeKind.USER
        ):
            raise ValidationError(input_value=f"{node_schema.kind}.{target_attribute.name} is not a computed attribute")

        context.active_permissions.raise_for_permission(
            permission=ObjectPermission(
                namespace=node_schema.namespace,
                name=node_schema.name,
                action=PermissionAction.UPDATE.value,
                decision=PermissionDecision.ALLOW_DEFAULT.value
                if context.branch.name == registry.default_branch
                else PermissionDecision.ALLOW_OTHER.value,
            )
        )

        if not (
            target_node := await NodeManager.get_one(
                db=context.db, kind=node_schema.kind, id=str(data.id), branch=context.branch
            )
        ):
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated node was not found in the database",
            )

        attribute_field = getattr(target_node, str(data.attribute), None)
        if not attribute_field:
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated not does not have the specified attribute_name",
            )
        if attribute_field.value != str(data.value):
            attribute_field.value = str(data.value)
            async with context.db.start_transaction() as dbt:
                await target_node.save(db=dbt, fields=[str(data.attribute)])

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            graphql_payload = await target_node.to_graphql(
                db=context.db, filter_sensitive=True, include_properties=False
            )

            event = NodeMutatedEvent(
                branch=context.branch.name,
                kind=node_schema.kind,
                node_id=target_node.get_id(),
                data=graphql_payload,
                fields=[str(data.attribute)],
                action=MutationAction.UPDATED,
                meta=EventMeta(initiator_id=WORKER_IDENTITY, request_id=request_id),
            )
            await context.active_service.event.send(event=event)

        result: dict[str, Any] = {"ok": True}

        return cls(**result)



from __future__ import annotations

from typing import Any

from pydantic import BaseModel

VARIABLE_TYPE_MAPPING = ((str, "String!"), (int, "Int!"), (float, "Float!"), (bool, "Boolean!"))


def convert_to_graphql_as_string(value: str | bool | list) -> str:
    if isinstance(value, str) and value.startswith("$"):
        return value
    if isinstance(value, str):
        return f'"{value}"'
    if isinstance(value, bool):
        return repr(value).lower()
    if isinstance(value, list):
        values_as_string = [convert_to_graphql_as_string(item) for item in value]
        return "[" + ", ".join(values_as_string) + "]"
    if isinstance(value, BaseModel):
        data = value.model_dump()
        return "{ " + ", ".join(f"{key}: {convert_to_graphql_as_string(val)}" for key, val in data.items()) + " }"

    return str(value)


def render_variables_to_string(data: dict[str, type[str | int | float | bool]]) -> str:
    """Render a dict into a variable string that will be used in a GraphQL Query.

    The $ sign will be automatically added to the name of the query.
    """
    vars_dict = {}
    for key, value in data.items():
        for class_type, var_string in VARIABLE_TYPE_MAPPING:
            if value == class_type:
                vars_dict[f"${key}"] = var_string

    return ", ".join([f"{key}: {value}" for key, value in vars_dict.items()])


def render_query_block(data: dict, offset: int = 4, indentation: int = 4) -> list[str]:
    FILTERS_KEY = "@filters"
    ALIAS_KEY = "@alias"
    KEYWORDS_TO_SKIP = [FILTERS_KEY, ALIAS_KEY]

    offset_str = " " * offset
    lines = []
    for key, value in data.items():
        if key in KEYWORDS_TO_SKIP:
            continue
        if value is None:
            lines.append(f"{offset_str}{key}")
        elif isinstance(value, dict) and len(value) == 1 and ALIAS_KEY in value and value[ALIAS_KEY]:
            lines.append(f"{offset_str}{value[ALIAS_KEY]}: {key}")
        elif isinstance(value, dict):
            if value.get(ALIAS_KEY):
                key_str = f"{value[ALIAS_KEY]}: {key}"
            else:
                key_str = key

            if value.get(FILTERS_KEY):
                filters_str = ", ".join(
                    [f"{key2}: {convert_to_graphql_as_string(value2)}" for key2, value2 in value[FILTERS_KEY].items()]
                )
                lines.append(f"{offset_str}{key_str}({filters_str}) " + "{")
            else:
                lines.append(f"{offset_str}{key_str} " + "{")

            lines.extend(render_query_block(data=value, offset=offset + indentation, indentation=indentation))
            lines.append(offset_str + "}")

    return lines


def render_input_block(data: dict, offset: int = 4, indentation: int = 4) -> list[str]:
    offset_str = " " * offset
    lines = []
    for key, value in data.items():
        if isinstance(value, dict):
            lines.append(f"{offset_str}{key}: " + "{")
            lines.extend(render_input_block(data=value, offset=offset + indentation, indentation=indentation))
            lines.append(offset_str + "}")
        elif isinstance(value, list):
            lines.append(f"{offset_str}{key}: " + "[")
            for item in value:
                if isinstance(item, dict):
                    lines.append(f"{offset_str}{' ' * indentation}" + "{")
                    lines.extend(
                        render_input_block(
                            data=item,
                            offset=offset + indentation + indentation,
                            indentation=indentation,
                        )
                    )
                    lines.append(f"{offset_str}{' ' * indentation}" + "},")
                else:
                    lines.append(f"{offset_str}{' ' * indentation}{convert_to_graphql_as_string(item)},")
            lines.append(offset_str + "]")
        else:
            lines.append(f"{offset_str}{key}: {convert_to_graphql_as_string(value)}")
    return lines


class BaseGraphQLQuery:
    query_type: str = "not-defined"
    indentation: int = 4

    def __init__(self, query: dict, variables: dict | None = None, name: str | None = None):
        self.query = query
        self.variables = variables
        self.name = name or ""

    def render_first_line(self) -> str:
        first_line = self.query_type

        if self.name:
            first_line += " " + self.name

        if self.variables:
            first_line += f" ({render_variables_to_string(self.variables)})"

        first_line += " {"

        return first_line


class Query(BaseGraphQLQuery):
    query_type = "query"

    def render(self) -> str:
        lines = [self.render_first_line()]
        lines.extend(render_query_block(data=self.query, indentation=self.indentation, offset=self.indentation))
        lines.append("}")

        return "\n" + "\n".join(lines) + "\n"


class Mutation(BaseGraphQLQuery):
    query_type = "mutation"

    def __init__(self, *args: Any, mutation: str, input_data: dict, **kwargs: Any):
        self.input_data = input_data
        self.mutation = mutation
        super().__init__(*args, **kwargs)

    def render(self) -> str:
        lines = [self.render_first_line()]
        lines.append(" " * self.indentation + f"{self.mutation}(")
        lines.extend(
            render_input_block(
                data=self.input_data,
                indentation=self.indentation,
                offset=self.indentation * 2,
            )
        )
        lines.append(" " * self.indentation + "){")
        lines.extend(
            render_query_block(
                data=self.query,
                indentation=self.indentation,
                offset=self.indentation * 2,
            )
        )
        lines.append(" " * self.indentation + "}")
        lines.append("}")

        return "\n" + "\n".join(lines) + "\n"
